# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Console for interacting with devices using HDLC.

To start the console, provide a serial port as the --device argument and paths
or globs for .proto files that define the RPC services to support:

  python -m pw_system.console --device /dev/ttyUSB0 --proto-globs pw_rpc/echo.proto

This starts an interactive console for communicating with the connected device.
A few variables are predefined in the console. These include:

    rpcs   - used to invoke RPCs
    device - the serial device used for communication
    client - the pw_rpc.Client
    protos - protocol buffer messages indexed by proto package

An example echo RPC command:

  rpcs.pw.rpc.EchoService.Echo(msg="hello!")
"""  # pylint: disable=line-too-long

import argparse
from inspect import cleandoc
import logging
from pathlib import Path
import sys
from types import ModuleType
from typing import (
    Callable,
    Collection,
    TYPE_CHECKING,
)

from pw_cli import log as pw_cli_log
from pw_console import embed
from pw_console import web
from pw_console import log_store
from pw_console.plugins import bandwidth_toolbar
from pw_console import python_logging
from pw_hdlc import rpc
from pw_rpc.console_tools.console import flattened_rpc_completions

from pw_system.device_connection import (
    DeviceConnection,
    add_device_args,
    create_device_serial_or_socket_connection,
)

if TYPE_CHECKING:
    from pw_system import device as pw_device
    from pw_system import device_tracing as pw_device_tracing

_LOG = logging.getLogger(__package__)
_DEVICE_LOG = logging.getLogger('rpc_device')
_SERIAL_DEBUG = logging.getLogger('pw_console.serial_debug_logger')
_ROOT_LOG = logging.getLogger()

MKFIFO_MODE = 0o666


def add_logfile_args(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Add logfile specific arguments needed by the interactive console."""

    def log_level(arg: str) -> int:
        try:
            return getattr(logging, arg.upper())
        except AttributeError:
            raise argparse.ArgumentTypeError(
                f'{arg.upper()} is not a valid log level'
            )

    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='Set the host log level to debug.',
    )
    parser.add_argument(
        '--host-log-level',
        type=log_level,
        default=logging.INFO,
        help='Set the host log level (debug, info, warning, error, critical)',
    )
    parser.add_argument(
        '--device-log-level',
        type=log_level,
        default=logging.DEBUG,
        help='Set the device log level (debug, info, warning, error, critical)',
    )
    parser.add_argument(
        '--json-logfile', help='Device only JSON formatted log file.'
    )
    parser.add_argument(
        '--logfile',
        default='pw_console-logs.txt',
        help=(
            'Default log file. This will contain host side '
            'log messages only unles the '
            '--merge-device-and-host-logs argument is used.'
        ),
    )
    parser.add_argument(
        '--merge-device-and-host-logs',
        action='store_true',
        help=(
            'Include device logs in the default --logfile.'
            'These are normally shown in a separate device '
            'only log file.'
        ),
    )
    parser.add_argument(
        '--host-logfile',
        help=(
            'Additional host only log file. Normally all logs in the '
            'default logfile are host only.'
        ),
    )
    parser.add_argument(
        '--device-logfile',
        default='pw_console-device-logs.txt',
        help='Device only log file.',
    )
    return parser


def add_pw_console_args(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Add pw_console startup specific arguments."""
    parser.add_argument(
        '--config-file',
        type=Path,
        help='Path to a pw_console yaml config file.',
    )

    parser.add_argument(
        "--browser",
        action="store_true",
        help="Start browser-based console instead of terminal.",
    )
    return parser


def get_parser() -> argparse.ArgumentParser:
    """Build pw_system.console argparse with both device and logfile args."""
    parser = argparse.ArgumentParser(
        prog='python -m pw_system.console',
        description=__doc__,
    )
    parser = add_device_args(parser)
    parser = add_logfile_args(parser)
    parser = add_pw_console_args(parser)
    return parser


def _parse_args(args: argparse.Namespace | None = None):
    """Parses and returns the command line arguments."""
    if args is not None:
        return args

    parser = get_parser()
    return parser.parse_args()


def _start_python_terminal(  # pylint: disable=too-many-arguments
    device: 'pw_device.Device',
    device_log_store: log_store.LogStore,
    root_log_store: log_store.LogStore,
    serial_debug_log_store: log_store.LogStore,
    log_file: str,
    host_logfile: str,
    device_logfile: str,
    json_logfile: str,
    serial_debug: bool = False,
    config_file_path: Path | None = None,
    browser: bool = False,
) -> None:
    """Starts an interactive Python terminal with preset variables."""
    local_variables = dict(
        client=device.client,
        device=device,
        rpcs=device.rpcs,
        protos=device.client.protos.packages,
        # Include the active pane logger for creating logs in the repl.
        DEVICE_LOG=_DEVICE_LOG,
        LOG=logging.getLogger(),
    )

    welcome_message = cleandoc(
        """
        Welcome to the Pigweed Console!

        Help: Press F1 or click the [Help] menu
        To move focus: Press Shift-Tab or click on a window

        Example Python commands:

          device.rpcs.pw.rpc.EchoService.Echo(msg='hello!')
          LOG.warning('Message appears in Host Logs window.')
          DEVICE_LOG.warning('Message appears in Device Logs window.')
    """
    )

    welcome_message += '\n\nLogs are being saved to:\n  ' + log_file
    if host_logfile:
        welcome_message += '\nHost logs are being saved to:\n  ' + host_logfile
    if device_logfile:
        welcome_message += (
            '\nDevice logs are being saved to:\n  ' + device_logfile
        )
    if json_logfile:
        welcome_message += (
            '\nJSON device logs are being saved to:\n  ' + json_logfile
        )

    client_info = device.info()
    completions = flattened_rpc_completions([client_info])

    log_windows: dict[str, list[logging.Logger] | log_store.LogStore] = {
        'Device Logs': device_log_store,
        'Host Logs': root_log_store,
    }
    if serial_debug:
        log_windows['Serial Debug'] = serial_debug_log_store

    if browser:
        loggers: dict[str, list[logging.Logger]] = {
            "Device Logs": [_DEVICE_LOG],
            "Host Logs": [_ROOT_LOG],
        }
        if serial_debug:
            loggers['Serial Debug'] = [_SERIAL_DEBUG]

        webserver = web.PwConsoleWeb(
            global_vars=local_variables,
            loggers=loggers,
            sentence_completions=completions,
        )
        webserver.start()
    else:
        interactive_console = embed.PwConsoleEmbed(
            global_vars=local_variables,
            local_vars=None,
            loggers=log_windows,
            repl_startup_message=welcome_message,
            help_text=__doc__,
            config_file_path=config_file_path,
        )
        interactive_console.add_sentence_completer(completions)
        if serial_debug:
            interactive_console.add_bottom_toolbar(
                bandwidth_toolbar.BandwidthToolbar()
            )

        # Setup Python logger propagation
        interactive_console.setup_python_logging(
            # Send any unhandled log messages to the external file.
            last_resort_filename=log_file,
            # Don't change propagation for these loggers.
            loggers_with_no_propagation=[_DEVICE_LOG],
        )

        interactive_console.embed()


# pylint: disable=too-many-arguments,too-many-locals
def console(
    device: str,
    baudrate: int,
    ticks_per_second: int | None,
    token_databases: Collection[Path],
    socket_addr: str,
    logfile: str,
    host_logfile: str,
    device_logfile: str,
    json_logfile: str,
    serial_debug: bool = False,
    config_file: Path | None = None,
    verbose: bool = False,
    host_log_level: int = logging.INFO,
    device_log_level: int = logging.DEBUG,
    compiled_protos: list[ModuleType] | None = None,
    merge_device_and_host_logs: bool = False,
    rpc_logging: bool = True,
    channel_id: int = rpc.DEFAULT_CHANNEL_ID,
    hdlc_encoding: bool = True,
    device_tracing: bool = True,
    browser: bool = False,
    timestamp_decoder: Callable[[int], str] | None = None,
    device_connection: DeviceConnection | None = None,
) -> int:
    """Starts an interactive RPC console for HDLC."""
    # Don't send device logs to the root logger.
    _DEVICE_LOG.propagate = False
    # Create pw_console log_store.LogStore handlers. These are the data source
    # for log messages to be displayed in the UI.
    device_log_store = log_store.LogStore()
    root_log_store = log_store.LogStore()
    serial_debug_log_store = log_store.LogStore()
    # Attach the log_store.LogStores as handlers for each log window we want to
    # show. This should be done before device initialization to capture early
    # messages.
    _DEVICE_LOG.addHandler(device_log_store)
    _ROOT_LOG.addHandler(root_log_store)
    _SERIAL_DEBUG.addHandler(serial_debug_log_store)

    if not logfile:
        # Create a temp logfile to prevent logs from appearing over stdout. This
        # would corrupt the prompt toolkit UI.
        logfile = python_logging.create_temp_log_file()

    if verbose:
        host_log_level = logging.DEBUG

    pw_cli_log.install(
        level=host_log_level,
        use_color=False,
        hide_timestamp=False,
        log_file=logfile,
    )

    if device_logfile:
        pw_cli_log.install(
            level=device_log_level,
            use_color=False,
            hide_timestamp=False,
            log_file=device_logfile,
            logger=_DEVICE_LOG,
        )
    if host_logfile:
        pw_cli_log.install(
            level=host_log_level,
            use_color=False,
            hide_timestamp=False,
            log_file=host_logfile,
            logger=_ROOT_LOG,
        )

    if merge_device_and_host_logs:
        # Add device logs to the default logfile.
        pw_cli_log.install(
            level=device_log_level,
            use_color=False,
            hide_timestamp=False,
            log_file=logfile,
            logger=_DEVICE_LOG,
        )

    _LOG.setLevel(host_log_level)
    _DEVICE_LOG.setLevel(device_log_level)
    _ROOT_LOG.setLevel(host_log_level)
    _SERIAL_DEBUG.setLevel(logging.DEBUG)

    if json_logfile:
        json_filehandler = logging.FileHandler(json_logfile, encoding='utf-8')
        json_filehandler.setLevel(device_log_level)
        json_filehandler.setFormatter(python_logging.JsonLogFormatter())
        _DEVICE_LOG.addHandler(json_filehandler)

    if device_connection is None:
        device_connection = create_device_serial_or_socket_connection(
            device=device,
            baudrate=baudrate,
            token_databases=token_databases,
            socket_addr=socket_addr,
            ticks_per_second=ticks_per_second,
            serial_debug=serial_debug,
            compiled_protos=compiled_protos,
            rpc_logging=rpc_logging,
            channel_id=channel_id,
            hdlc_encoding=hdlc_encoding,
            device_tracing=device_tracing,
            timestamp_decoder=timestamp_decoder,
        )

    with device_connection as device_client:
        _start_python_terminal(
            device=device_client,
            device_log_store=device_log_store,
            root_log_store=root_log_store,
            serial_debug_log_store=serial_debug_log_store,
            log_file=logfile,
            host_logfile=host_logfile,
            device_logfile=device_logfile,
            json_logfile=json_logfile,
            serial_debug=serial_debug,
            config_file_path=config_file,
            browser=browser,
        )
    return 0


def main(
    args: argparse.Namespace | None = None,
    compiled_protos: list[ModuleType] | None = None,
    timestamp_decoder: Callable[[int], str] | None = None,
    device_connection: DeviceConnection | None = None,
) -> int:
    """Startup the pw console UI for a pw_system device.

    Example usage:

    .. code-block:: python

       from pw_protobuf_protos import common_pb2
       from pw_rpc import echo_pb2
       from pw_log.log_decoder.timestamp_parser_ms_since_boot
       import pw_system.console

       def main() -> int:
           return pw_system.console.main(
               compiled_protos=[
                   common_pb2,
                   echo_pb2,
               ],
               timestamp_decoder=timestamp_parser_ms_since_boot,
               device_connection=None,
           )

       if __name__ == '__main__':
           sys.exit(main())
    """
    return console(
        **vars(_parse_args(args)),
        compiled_protos=compiled_protos,
        timestamp_decoder=timestamp_decoder,
        device_connection=device_connection,
    )


# TODO(b/353327855): Deprecated function, remove when no longer used.
def main_with_compiled_protos(
    compiled_protos: list[ModuleType] | None = None,
    args: argparse.Namespace | None = None,
    device_connection: DeviceConnection | None = None,
) -> int:
    return main(
        args=args,
        compiled_protos=compiled_protos,
        device_connection=device_connection,
    )


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_transfer/transfer.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1apw_transfer/transfer.proto\x12\x0bpw.transfer\"\xfd\x05\n\x05\x43hunk\x12\x13\n\x0btransfer_id\x18\x01 \x01(\r\x12\x1a\n\rpending_bytes\x18\x02 \x01(\rH\x00\x88\x01\x01\x12!\n\x14max_chunk_size_bytes\x18\x03 \x01(\rH\x01\x88\x01\x01\x12#\n\x16min_delay_microseconds\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x0e\n\x06offset\x18\x05 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\x1c\n\x0fremaining_bytes\x18\x07 \x01(\x04H\x03\x88\x01\x01\x12\x13\n\x06status\x18\x08 \x01(\rH\x04\x88\x01\x01\x12\x19\n\x11window_end_offset\x18\t \x01(\r\x12*\n\x04type\x18\n \x01(\x0e\x32\x17.pw.transfer.Chunk.TypeH\x05\x88\x01\x01\x12\x18\n\x0bresource_id\x18\x0b \x01(\rH\x06\x88\x01\x01\x12\x17\n\nsession_id\x18\x0c \x01(\rH\x07\x88\x01\x01\x12\x1d\n\x10protocol_version\x18\r \x01(\rH\x08\x88\x01\x01\x12\x1f\n\x12\x64\x65sired_session_id\x18\x0e \x01(\rH\t\x88\x01\x01\x12\x16\n\x0einitial_offset\x18\x0f \x01(\x04\"\x9e\x01\n\x04Type\x12\x08\n\x04\x44\x41TA\x10\x00\x12\t\n\x05START\x10\x01\x12\x19\n\x15PARAMETERS_RETRANSMIT\x10\x02\x12\x17\n\x13PARAMETERS_CONTINUE\x10\x03\x12\x0e\n\nCOMPLETION\x10\x04\x12\x12\n\x0e\x43OMPLETION_ACK\x10\x05\x12\r\n\tSTART_ACK\x10\x06\x12\x1a\n\x16START_ACK_CONFIRMATION\x10\x07\x42\x10\n\x0e_pending_bytesB\x17\n\x15_max_chunk_size_bytesB\x19\n\x17_min_delay_microsecondsB\x12\n\x10_remaining_bytesB\t\n\x07_statusB\x07\n\x05_typeB\x0e\n\x0c_resource_idB\r\n\x0b_session_idB\x13\n\x11_protocol_versionB\x15\n\x13_desired_session_id\",\n\x15ResourceStatusRequest\x12\x13\n\x0bresource_id\x18\x01 \x01(\r\"\xc6\x01\n\x0eResourceStatus\x12\x13\n\x0bresource_id\x18\x01 \x01(\r\x12\x0e\n\x06status\x18\x02 \x01(\r\x12\x18\n\x10writeable_offset\x18\x03 \x01(\x04\x12\x17\n\x0freadable_offset\x18\x04 \x01(\x04\x12\x1b\n\x0ewrite_checksum\x18\x05 \x01(\x04H\x00\x88\x01\x01\x12\x1a\n\rread_checksum\x18\x06 \x01(\x04H\x01\x88\x01\x01\x42\x11\n\x0f_write_checksumB\x10\n\x0e_read_checksum2\xc9\x01\n\x08Transfer\x12\x32\n\x04Read\x12\x12.pw.transfer.Chunk\x1a\x12.pw.transfer.Chunk(\x01\x30\x01\x12\x33\n\x05Write\x12\x12.pw.transfer.Chunk\x1a\x12.pw.transfer.Chunk(\x01\x30\x01\x12T\n\x11GetResourceStatus\x12\".pw.transfer.ResourceStatusRequest\x1a\x1b.pw.transfer.ResourceStatusB\x1b\n\x17\x64\x65v.pigweed.pw_transferP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_transfer.transfer_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\027dev.pigweed.pw_transferP\001'
  _globals['_CHUNK']._serialized_start=44
  _globals['_CHUNK']._serialized_end=809
  _globals['_CHUNK_TYPE']._serialized_start=466
  _globals['_CHUNK_TYPE']._serialized_end=624
  _globals['_RESOURCESTATUSREQUEST']._serialized_start=811
  _globals['_RESOURCESTATUSREQUEST']._serialized_end=855
  _globals['_RESOURCESTATUS']._serialized_start=858
  _globals['_RESOURCESTATUS']._serialized_end=1056
  _globals['_TRANSFER']._serialized_start=1059
  _globals['_TRANSFER']._serialized_end=1260
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/benchmark.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16pw_rpc/benchmark.proto\x12\x06pw.rpc\"\x1a\n\x07Payload\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x32u\n\tBenchmark\x12-\n\tUnaryEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload\x12\x39\n\x11\x42idirectionalEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_rpc.benchmark_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PAYLOAD']._serialized_start=34
  _globals['_PAYLOAD']._serialized_end=60
  _globals['_BENCHMARK']._serialized_start=62
  _globals['_BENCHMARK']._serialized_end=179
# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""The callback-based pw_rpc client implementation."""

from __future__ import annotations

import inspect
import logging
import textwrap
from typing import Any, Callable, Iterable, Type

from dataclasses import dataclass
from pw_status import Status
from google.protobuf.message import Message

from pw_rpc import client, descriptors
from pw_rpc.client import PendingRpc, PendingRpcs
from pw_rpc.descriptors import Channel, Method, Service

from pw_rpc.callback_client.call import (
    UseDefault,
    OptionalTimeout,
    CallTypeT,
    UnaryResponse,
    StreamResponse,
    Call,
    UnaryCall,
    ServerStreamingCall,
    ClientStreamingCall,
    BidirectionalStreamingCall,
    OnNextCallback,
    OnCompletedCallback,
    OnErrorCallback,
)

_LOG = logging.getLogger(__package__)


DEFAULT_MAX_STREAM_RESPONSES = 2**14


@dataclass(eq=True, frozen=True)
class CallInfo:
    method: Method

    @property
    def service(self) -> Service:
        return self.method.service


class _MethodClient:
    """A method that can be invoked for a particular channel."""

    def __init__(
        self,
        client_impl: Impl,
        rpcs: PendingRpcs,
        channel: Channel,
        method: Method,
        default_timeout_s: float | None,
    ) -> None:
        self._impl = client_impl
        self._rpcs = rpcs
        self._channel = channel
        self._method = method
        self.default_timeout_s: float | None = default_timeout_s

    @property
    def channel(self) -> Channel:
        return self._channel

    @property
    def method(self) -> Method:
        return self._method

    @property
    def service(self) -> Service:
        return self._method.service

    @property
    def request(self) -> type:
        """Returns the request proto class."""
        return self.method.request_type

    @property
    def response(self) -> type:
        """Returns the response proto class."""
        return self.method.response_type

    def __repr__(self) -> str:
        return self.help()

    def help(self) -> str:
        """Returns a help message about this RPC."""
        function_call = self.method.full_name + '('

        docstring = inspect.getdoc(self.__call__)  # type: ignore[operator] # pylint: disable=no-member
        assert docstring is not None

        annotation = inspect.signature(self).return_annotation  # type: ignore[arg-type] # pylint: disable=line-too-long
        if isinstance(annotation, type):
            annotation = annotation.__name__

        arg_sep = f',\n{" " * len(function_call)}'
        return (
            f'{function_call}'
            f'{arg_sep.join(descriptors.field_help(self.method.request_type))})'
            f'\n\n{textwrap.indent(docstring, "  ")}\n\n'
            f'  Returns {annotation}.'
        )

    def _start_call(
        self,
        call_type: Type[CallTypeT],
        request: Message | None,
        timeout_s: OptionalTimeout,
        on_next: OnNextCallback | None,
        on_completed: OnCompletedCallback | None,
        on_error: OnErrorCallback | None,
        max_responses: int,
    ) -> CallTypeT:
        """Creates the Call object and invokes the RPC using it."""
        if timeout_s is UseDefault.VALUE:
            timeout_s = self.default_timeout_s

        if self._impl.on_call_hook:
            self._impl.on_call_hook(CallInfo(self._method))

        rpc = PendingRpc(
            self._channel,
            self.service,
            self.method,
            self._rpcs.allocate_call_id(),
        )
        call = call_type(
            self._rpcs,
            rpc,
            timeout_s,
            on_next,
            on_completed,
            on_error,
            max_responses,
        )
        call._invoke(request)  # pylint: disable=protected-access
        return call

    def _open_call(
        self,
        call_type: Type[CallTypeT],
        on_next: OnNextCallback | None,
        on_completed: OnCompletedCallback | None,
        on_error: OnErrorCallback | None,
        max_responses: int,
    ) -> CallTypeT:
        """Creates a Call object with the open call ID."""
        rpc = PendingRpc(
            self._channel,
            self.service,
            self.method,
            client.OPEN_CALL_ID,
        )
        call = call_type(
            self._rpcs,
            rpc,
            None,
            on_next,
            on_completed,
            on_error,
            max_responses,
        )
        call._open()  # pylint: disable=protected-access
        return call

    def _client_streaming_call_type(
        self, base: Type[CallTypeT]
    ) -> Type[CallTypeT]:
        """Creates a client or bidirectional stream call type.

        Applies the signature from the request protobuf to the send method.
        """

        def send(
            self, request_proto: Message | None = None, /, **request_fields
        ) -> None:
            ClientStreamingCall.send(self, request_proto, **request_fields)

        _apply_protobuf_signature(self.method, send)

        return type(
            f'{self.method.name}_{base.__name__}', (base,), dict(send=send)
        )


def _function_docstring(method: Method) -> str:
    return f'''\
Invokes the {method.full_name} {method.type.sentence_name()} RPC.

This function accepts either the request protobuf fields as keyword arguments or
a request protobuf as a positional argument.
'''


def _update_call_method(method: Method, function: Callable) -> None:
    """Updates the name, docstring, and parameters to match a method."""
    function.__name__ = method.full_name
    function.__doc__ = _function_docstring(method)
    _apply_protobuf_signature(method, function)


def _apply_protobuf_signature(method: Method, function: Callable) -> None:
    """Update a function signature to accept proto arguments.

    In order to have good tab completion and help messages, update the function
    signature to accept only keyword arguments for the proto message fields.
    This doesn't actually change the function signature -- it just updates how
    it appears when inspected.
    """
    sig = inspect.signature(function)

    params = [next(iter(sig.parameters.values()))]  # Get the "self" parameter
    params += method.request_parameters()
    params.append(
        inspect.Parameter('pw_rpc_timeout_s', inspect.Parameter.KEYWORD_ONLY)
    )

    function.__signature__ = sig.replace(  # type: ignore[attr-defined]
        parameters=params
    )


class _UnaryMethodClient(_MethodClient):
    def invoke(
        self,
        request: Message | None = None,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
        *,
        request_args: dict[str, Any] | None = None,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> UnaryCall:
        """Invokes the unary RPC and returns a call object."""
        return self._start_call(
            UnaryCall,
            self.method.get_request(request, request_args),
            timeout_s,
            on_next,
            on_completed,
            on_error,
            max_responses=1,
        )

    def open(
        self,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
    ) -> UnaryCall:
        """Invokes the unary RPC and returns a call object."""
        return self._open_call(
            UnaryCall, on_next, on_completed, on_error, max_responses=1
        )


class _ServerStreamingMethodClient(_MethodClient):
    def invoke(
        self,
        request: Message | None = None,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
        max_responses: int = DEFAULT_MAX_STREAM_RESPONSES,
        *,
        request_args: dict[str, Any] | None = None,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> ServerStreamingCall:
        """Invokes the server streaming RPC and returns a call object."""
        return self._start_call(
            ServerStreamingCall,
            self.method.get_request(request, request_args),
            timeout_s,
            on_next,
            on_completed,
            on_error,
            max_responses=max_responses,
        )

    def open(
        self,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
        max_responses: int = DEFAULT_MAX_STREAM_RESPONSES,
    ) -> ServerStreamingCall:
        """Returns a call object for the RPC, even if the RPC cannot be invoked.

        Can be used to listen for responses from an RPC server that may yet be
        available.
        """
        return self._open_call(
            ServerStreamingCall, on_next, on_completed, on_error, max_responses
        )


class _ClientStreamingMethodClient(_MethodClient):
    def invoke(
        self,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> ClientStreamingCall:
        """Invokes the client streaming RPC and returns a call object"""
        return self._start_call(
            self._client_streaming_call_type(ClientStreamingCall),
            None,
            timeout_s,
            on_next,
            on_completed,
            on_error,
            max_responses=1,
        )

    def open(
        self,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
    ) -> ClientStreamingCall:
        """Returns a call object for the RPC, even if the RPC cannot be invoked.

        Can be used to listen for responses from an RPC server that may yet be
        available.
        """
        return self._open_call(
            self._client_streaming_call_type(ClientStreamingCall),
            on_next,
            on_completed,
            on_error,
            max_responses=1,
        )

    def __call__(
        self,
        requests: Iterable[Message] = (),
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> UnaryResponse:
        return self.invoke().finish_and_wait(requests, timeout_s=timeout_s)


class _BidirectionalStreamingMethodClient(_MethodClient):
    def invoke(
        self,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
        max_responses: int = DEFAULT_MAX_STREAM_RESPONSES,
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> BidirectionalStreamingCall:
        """Invokes the bidirectional streaming RPC and returns a call object."""
        return self._start_call(
            self._client_streaming_call_type(BidirectionalStreamingCall),
            None,
            timeout_s,
            on_next,
            on_completed,
            on_error,
            max_responses=max_responses,
        )

    def open(
        self,
        on_next: OnNextCallback | None = None,
        on_completed: OnCompletedCallback | None = None,
        on_error: OnErrorCallback | None = None,
        max_responses: int = DEFAULT_MAX_STREAM_RESPONSES,
    ) -> BidirectionalStreamingCall:
        """Returns a call object for the RPC, even if the RPC cannot be invoked.

        Can be used to listen for responses from an RPC server that may yet be
        available.
        """
        return self._open_call(
            self._client_streaming_call_type(BidirectionalStreamingCall),
            on_next,
            on_completed,
            on_error,
            max_responses=max_responses,
        )

    def __call__(
        self,
        requests: Iterable[Message] = (),
        *,
        timeout_s: OptionalTimeout = UseDefault.VALUE,
    ) -> StreamResponse:
        return self.invoke().finish_and_wait(requests, timeout_s=timeout_s)


def _method_client_docstring(method: Method) -> str:
    return f'''\
Class that invokes the {method.full_name} {method.type.sentence_name()} RPC.

Calling this directly invokes the RPC synchronously. The RPC can be invoked
asynchronously using the invoke method.
'''


class Impl(client.ClientImpl):
    """Callback-based ClientImpl, for use with pw_rpc.Client.

    Args:
        on_call_hook: A callable object to handle RPC method calls.
            If hook is set, it will be called before RPC execution.
    """

    def __init__(
        self,
        default_unary_timeout_s: float | None = None,
        default_stream_timeout_s: float | None = None,
        on_call_hook: Callable[[CallInfo], Any] | None = None,
    ) -> None:
        super().__init__()
        self._default_unary_timeout_s = default_unary_timeout_s
        self._default_stream_timeout_s = default_stream_timeout_s
        self.on_call_hook = on_call_hook

    @property
    def default_unary_timeout_s(self) -> float | None:
        return self._default_unary_timeout_s

    @property
    def default_stream_timeout_s(self) -> float | None:
        return self._default_stream_timeout_s

    def method_client(self, channel: Channel, method: Method) -> _MethodClient:
        """Returns an object that invokes a method using the given chanel."""

        if method.type is Method.Type.UNARY:
            return self._create_unary_method_client(
                channel, method, self.default_unary_timeout_s
            )

        if method.type is Method.Type.SERVER_STREAMING:
            return self._create_server_streaming_method_client(
                channel, method, self.default_stream_timeout_s
            )

        if method.type is Method.Type.CLIENT_STREAMING:
            return self._create_method_client(
                _ClientStreamingMethodClient,
                channel,
                method,
                self.default_unary_timeout_s,
            )

        if method.type is Method.Type.BIDIRECTIONAL_STREAMING:
            return self._create_method_client(
                _BidirectionalStreamingMethodClient,
                channel,
                method,
                self.default_stream_timeout_s,
            )

        raise AssertionError(f'Unknown method type {method.type}')

    def _create_method_client(
        self,
        base: type,
        channel: Channel,
        method: Method,
        default_timeout_s: float | None,
        **fields,
    ):
        """Creates a _MethodClient derived class customized for the method."""
        method_client_type = type(
            f'{method.name}{base.__name__}',
            (base,),
            dict(__doc__=_method_client_docstring(method), **fields),
        )
        return method_client_type(
            self, self.rpcs, channel, method, default_timeout_s
        )

    def _create_unary_method_client(
        self,
        channel: Channel,
        method: Method,
        default_timeout_s: float | None,
    ) -> _UnaryMethodClient:
        """Creates a _UnaryMethodClient with a customized __call__ method."""

        def call(
            self: _UnaryMethodClient,
            request_proto: Message | None = None,
            /,
            *,
            pw_rpc_timeout_s: OptionalTimeout = UseDefault.VALUE,
            **request_fields,
        ) -> UnaryResponse:
            return self.invoke(
                self.method.get_request(request_proto, request_fields)
            ).wait(pw_rpc_timeout_s)

        _update_call_method(method, call)
        return self._create_method_client(
            _UnaryMethodClient,
            channel,
            method,
            default_timeout_s,
            __call__=call,
        )

    def _create_server_streaming_method_client(
        self,
        channel: Channel,
        method: Method,
        default_timeout_s: float | None,
    ) -> _ServerStreamingMethodClient:
        """Creates _ServerStreamingMethodClient with custom __call__ method."""

        def call(
            self: _ServerStreamingMethodClient,
            request_proto: Message | None = None,
            /,
            *,
            pw_rpc_timeout_s: OptionalTimeout = UseDefault.VALUE,
            **request_fields,
        ) -> StreamResponse:
            return self.invoke(
                self.method.get_request(request_proto, request_fields)
            ).wait(pw_rpc_timeout_s)

        _update_call_method(method, call)
        return self._create_method_client(
            _ServerStreamingMethodClient,
            channel,
            method,
            default_timeout_s,
            __call__=call,
        )

    def handle_response(
        self,
        rpc: PendingRpc,
        context: Call,
        payload,
    ) -> None:
        """Invokes the callback associated with this RPC."""
        context._handle_response(payload)  # pylint: disable=protected-access

    def handle_completion(
        self,
        rpc: PendingRpc,
        context: Call,
        status: Status,
    ):
        context._handle_completion(status)  # pylint: disable=protected-access

    def handle_error(
        self,
        rpc: PendingRpc,
        context: Call,
        status: Status,
    ) -> None:
        context._handle_error(status)  # pylint: disable=protected-access



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2023 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import pw_chrono_protos.chrono_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class StartRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___StartRequest = StartRequest

@typing_extensions.final
class StartResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___StartResponse = StartResponse

@typing_extensions.final
class StopRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___StopRequest = StopRequest

@typing_extensions.final
class StopResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_ID_FIELD_NUMBER: builtins.int
    file_id: builtins.int
    """as a convenience, the file id is returned on stop which can be
    used to start a transfer directly, rather that requiring a user
    list the files to obtain the file id.
    """
    def __init__(
        self,
        *,
        file_id: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_file_id", b"_file_id", "file_id", b"file_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_file_id", b"_file_id", "file_id", b"file_id"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_file_id", b"_file_id"]) -> typing_extensions.Literal["file_id"] | None: ...

global___StopResponse = StopResponse

@typing_extensions.final
class ClockParametersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ClockParametersRequest = ClockParametersRequest

@typing_extensions.final
class ClockParametersResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CLOCK_PARAMETERS_FIELD_NUMBER: builtins.int
    @property
    def clock_parameters(self) -> pw_chrono_protos.chrono_pb2.ClockParameters: ...
    def __init__(
        self,
        *,
        clock_parameters: pw_chrono_protos.chrono_pb2.ClockParameters | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["clock_parameters", b"clock_parameters"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["clock_parameters", b"clock_parameters"]) -> None: ...

global___ClockParametersResponse = ClockParametersResponse



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console entry point."""

import argparse
import inspect
import logging
from pathlib import Path
import sys

from pw_cli import log as pw_cli_log
from pw_cli import argument_types
from pw_console import PwConsoleEmbed
from pw_console.python_logging import create_temp_log_file
from pw_console.log_store import LogStore
from pw_console.plugins.calc_pane import CalcPane
from pw_console.plugins.clock_pane import ClockPane
from pw_console.plugins.twenty48_pane import Twenty48Pane
from pw_console.test_mode import FAKE_DEVICE_LOGGER_NAME
from pw_console.web import PwConsoleWeb

_LOG = logging.getLogger(__package__)
_ROOT_LOG = logging.getLogger('')


def _build_argument_parser() -> argparse.ArgumentParser:
    """Setup argparse."""
    parser = argparse.ArgumentParser(
        prog="python -m pw_console", description=__doc__
    )

    parser.add_argument(
        '-l',
        '--loglevel',
        type=argument_types.log_level,
        default=logging.DEBUG,
        help='Set the log level' '(debug, info, warning, error, critical)',
    )

    parser.add_argument('--logfile', help='Pigweed Console log file.')

    parser.add_argument(
        '--test-mode',
        action='store_true',
        help='Enable fake log messages for testing purposes.',
    )
    parser.add_argument(
        '--config-file',
        type=Path,
        help='Path to a pw_console yaml config file.',
    )
    parser.add_argument(
        '--console-debug-log-file',
        help='Log file to send console debug messages to.',
    )
    parser.add_argument(
        "--browser",
        action="store_true",
        help="Start browser-based console instead of terminal.",
    )

    return parser


def main(args: argparse.Namespace | None = None) -> int:
    """Pigweed Console."""

    parser = _build_argument_parser()

    if args is None:
        args = parser.parse_args()

    if not args.logfile:
        # Create a temp logfile to prevent logs from appearing over stdout. This
        # would corrupt the prompt toolkit UI.
        args.logfile = create_temp_log_file()

    pw_cli_log.install(
        level=args.loglevel,
        use_color=True,
        hide_timestamp=False,
        log_file=args.logfile,
    )

    if args.console_debug_log_file:
        pw_cli_log.install(
            level=logging.DEBUG,
            use_color=True,
            hide_timestamp=False,
            log_file=args.console_debug_log_file,
            logger=logging.getLogger('pw_console'),
        )

    global_vars = None
    default_loggers = {}
    if args.test_mode:
        root_log_store = LogStore()
        _ROOT_LOG.addHandler(root_log_store)
        _ROOT_LOG.debug('pw_console test-mode starting...')

        fake_logger = logging.getLogger(FAKE_DEVICE_LOGGER_NAME)
        default_loggers = {
            # Don't include pw_console package logs (_LOG) in the log pane UI.
            # Add the fake logger for test_mode.
            'Fake Device': [fake_logger],
            'PwConsole Debug': [logging.getLogger('pw_console')],
            'All Logs': root_log_store,
        }
        # Give access to adding log messages from the repl via: `LOG.warning()`
        global_vars = dict(LOG=fake_logger)

    help_text = None
    app_title = None
    if args.test_mode:
        app_title = 'Console Test Mode'
        help_text = inspect.cleandoc(
            """
            Welcome to the Pigweed Console Test Mode!

            Example commands:

              rpcs.pw.rpc.EchoService.Echo(msg='hello!')

              LOG.warning('Message appears console log window.')
        """
        )

    if args.browser:
        loggers = {
            'Fake Device': [fake_logger],
            'PwConsole Debug': [logging.getLogger('pw_console')],
            'All Logs': [_ROOT_LOG],
        }
        webserver = PwConsoleWeb(
            global_vars=global_vars,
            loggers=loggers,
            test_mode=args.test_mode,
        )
        webserver.start()
    else:
        console = PwConsoleEmbed(
            global_vars=global_vars,
            loggers=default_loggers,
            test_mode=args.test_mode,
            help_text=help_text,
            app_title=app_title,
            config_file_path=args.config_file,
        )

        overridden_window_config: dict | None = None
        # Add example plugins and log panes used to validate behavior in the
        # Pigweed Console manual test procedure:
        # https://pigweed.dev/pw_console/testing.html
        if args.test_mode:
            fake_logger.propagate = False
            console.setup_python_logging(
                loggers_with_no_propagation=[fake_logger]
            )

            _ROOT_LOG.debug("pw_console.PwConsoleEmbed init complete")
            _ROOT_LOG.debug("Adding plugins...")
            console.add_window_plugin(ClockPane())
            console.add_window_plugin(CalcPane())
            console.add_floating_window_plugin(
                Twenty48Pane(include_resize_handle=False), left=4
            )
            _ROOT_LOG.debug(
                "Starting prompt_toolkit full-screen application..."
            )

            overridden_window_config = {
                "Split 1 stacked": {
                    "Fake Device": None,
                    "Fake Keys": {
                        "duplicate_of": "Fake Device",
                        "filters": {
                            "keys": {"regex": "[^ ]+"},
                        },
                    },
                    "Fake USB": {
                        "duplicate_of": "Fake Device",
                        "filters": {
                            "module": {"regex": "USB"},
                        },
                    },
                },
                "Split 2 tabbed": {
                    "Python Repl": None,
                    "All Logs": None,
                    "PwConsole Debug": None,
                    "Calculator": None,
                    "Clock": None,
                },
            }

        console.embed(override_window_config=overridden_window_config)

    if args.logfile:
        print(f'Logs saved to: {args.logfile}')

    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/internal/packet.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cpw_rpc/internal/packet.proto\x12\x0fpw.rpc.internal\"\xa3\x01\n\tRpcPacket\x12)\n\x04type\x18\x01 \x01(\x0e\x32\x1b.pw.rpc.internal.PacketType\x12\x12\n\nchannel_id\x18\x02 \x01(\r\x12\x12\n\nservice_id\x18\x03 \x01(\x07\x12\x11\n\tmethod_id\x18\x04 \x01(\x07\x12\x0f\n\x07payload\x18\x05 \x01(\x0c\x12\x0e\n\x06status\x18\x06 \x01(\r\x12\x0f\n\x07\x63\x61ll_id\x18\x07 \x01(\r*\x9c\x01\n\nPacketType\x12\x0b\n\x07REQUEST\x10\x00\x12\x11\n\rCLIENT_STREAM\x10\x02\x12\x10\n\x0c\x43LIENT_ERROR\x10\x04\x12\x1d\n\x19\x43LIENT_REQUEST_COMPLETION\x10\x08\x12\x0c\n\x08RESPONSE\x10\x01\x12\x10\n\x0cSERVER_ERROR\x10\x05\x12\x11\n\rSERVER_STREAM\x10\x07\"\x04\x08\x03\x10\x03\"\x04\x08\x06\x10\x06\x42\x1d\n\x1b\x64\x65v.pigweed.pw_rpc.internalb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_rpc.internal.packet_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033dev.pigweed.pw_rpc.internal'
  _globals['_PACKETTYPE']._serialized_start=216
  _globals['_PACKETTYPE']._serialized_end=372
  _globals['_RPCPACKET']._serialized_start=50
  _globals['_RPCPACKET']._serialized_end=213
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw module template loading and files.

This directory should contain files ending in '.jinja'.

All files need to be added to:
  1. //pw_module/py/BUILD.gn as inputs
  2. //pw_module/py/setup.cfg under package_data

For jinja syntax help see: https://jinja.palletsprojects.com/en/3.1.x/templates/
"""

from importlib.resources import files
from itertools import chain
import logging

from jinja2 import Environment, DictLoader, make_logging_undefined
from jinja2.environment import Template

TEMPLATE_PATH = files('pw_module.templates')

# Dict containing file names and their contents.
JINJA_TEMPLATES = {
    t.relative_to(TEMPLATE_PATH).as_posix(): t.read_text()  # type: ignore
    for t in chain(
        files('pw_module.templates').iterdir(),
        files('pw_module.templates.helpers').iterdir(),
    )
    if t.suffix == '.jinja'  # type: ignore
}

JINJA_ENV = Environment(
    # Load templates automatically from pw_console/templates
    loader=DictLoader(JINJA_TEMPLATES),
    extensions=[
        # https://jinja.palletsprojects.com/en/3.1.x/extensions/#expression-statement
        'jinja2.ext.do',
    ],
    # Raise errors if variables are undefined in templates
    undefined=make_logging_undefined(
        logger=logging.getLogger(__package__),
    ),
    # Trim whitespace in templates
    trim_blocks=True,
    lstrip_blocks=True,
)


def get_template(file_name: str) -> Template:
    """Return the Template defined by a given file_name.

    Raises:
      jinja2.exceptions.TemplateNotFound when the template file name can't be
          found.
    """
    return JINJA_ENV.get_template(file_name)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Sphinx directives for Pigweed SEEDs"""


import docutils
from docutils import nodes
import docutils.statemachine

# pylint: disable=consider-using-from-import
import docutils.parsers.rst.directives as directives  # type: ignore

# pylint: enable=consider-using-from-import
from sphinx.application import Sphinx as SphinxApplication
from sphinx.util.docutils import SphinxDirective

from sphinx_design.cards import CardDirective


def status_choice(arg) -> str:
    return directives.choice(
        arg,
        (
            'draft',
            'open_for_comments',
            'intent_approved',
            'last_call',
            'accepted',
            'rejected',
        ),
    )


def parse_status(arg) -> str:
    """Support variations on the status choices.

    For example, you can use capital letters and spaces.
    """

    return status_choice('_'.join([token.lower() for token in arg.split(' ')]))


def status_badge(seed_status: str, badge_status) -> str:
    """Given a SEED status, return the status badge for rendering."""

    return (
        ':bdg-primary:'
        if seed_status == badge_status
        else ':bdg-secondary-line:'
    )


def cl_link(cl_num):
    return (
        f'`pwrev/{cl_num} '
        '<https://pigweed-review.googlesource.com/c/pigweed/pigweed/+/'
        f'{cl_num}>`_'
    )


class PigweedSeedDirective(SphinxDirective):
    """Directive registering & rendering SEED metadata."""

    required_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        'number': directives.positive_int,
        'name': directives.unchanged_required,
        'status': parse_status,
        'proposal_date': directives.unchanged_required,
        'cl': directives.positive_int_list,
        'authors': directives.unchanged_required,
        'facilitator': directives.unchanged_required,
    }

    def _try_get_option(self, option: str):
        """Try to get an option by name and raise on failure."""

        try:
            return self.options[option]
        except KeyError:
            raise self.error(f' :{option}: option is required')

    def run(self) -> list[nodes.Node]:
        seed_number = '{:04d}'.format(self._try_get_option('number'))
        seed_name = self._try_get_option('name')
        status = self._try_get_option('status')
        proposal_date = self._try_get_option('proposal_date')
        cl_nums = self._try_get_option('cl')
        authors = self._try_get_option('authors')
        facilitator = self._try_get_option('facilitator')

        title = (
            f':fas:`seedling` SEED-{seed_number}: :ref:'
            f'`{seed_name}<seed-{seed_number}>`\n'
        )

        authors_heading = 'Authors' if len(authors.split(',')) > 1 else 'Author'

        self.content = docutils.statemachine.StringList(
            [
                ':octicon:`comment-discussion` Status:',
                f'{status_badge(status, "open_for_comments")}'
                '`Open for Comments`',
                ':octicon:`chevron-right`',
                f'{status_badge(status, "intent_approved")}'
                '`Intent Approved`',
                ':octicon:`chevron-right`',
                f'{status_badge(status, "last_call")}`Last Call`',
                ':octicon:`chevron-right`',
                f'{status_badge(status, "accepted")}`Accepted`',
                ':octicon:`kebab-horizontal`',
                f'{status_badge(status, "rejected")}`Rejected`',
                '\n',
                f':octicon:`calendar` Proposal Date: {proposal_date}',
                '\n',
                ':octicon:`code-review` CL: ',
                ', '.join([cl_link(cl_num) for cl_num in cl_nums]),
                '\n',
                f':octicon:`person` {authors_heading}: {authors}',
                '\n',
                f':octicon:`person` Facilitator: {facilitator}',
            ]
        )

        card = CardDirective.create_card(
            inst=self,
            arguments=[title],
            options={},
        )

        return [card]


def setup(app: SphinxApplication):
    app.add_directive('seed', PigweedSeedDirective)

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Provides a simple interface for transferring bulk data over pw_rpc."""

# The generated protos for this module overlap this `__init__.py` file's import
# namespace, so we need to use extend_path() for them to be discoverable.
# Note: this needs to be done in every nested `__init__.py` file as well (if
# any exist).
__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from pw_transfer.transfer import (
    ProgressCallback,
    ProgressStats,
    ProtocolVersion,
)
from pw_transfer.client import Error, Manager



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_trace_protos/trace_rpc.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fpw_trace_protos/trace_rpc.proto\x12\x08pw.trace\"\x07\n\x05\x45mpty\"$\n\x12TraceEnableMessage\x12\x0e\n\x06\x65nable\x18\x01 \x01(\x08\" \n\x10TraceDataMessage\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x32\xd5\x01\n\x0cTraceService\x12\x46\n\x06\x45nable\x12\x1c.pw.trace.TraceEnableMessage\x1a\x1c.pw.trace.TraceEnableMessage\"\x00\x12<\n\tIsEnabled\x12\x0f.pw.trace.Empty\x1a\x1c.pw.trace.TraceEnableMessage\"\x00\x12?\n\x0cGetTraceData\x12\x0f.pw.trace.Empty\x1a\x1a.pw.trace.TraceDataMessage\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_trace_protos.trace_rpc_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EMPTY']._serialized_start=45
  _globals['_EMPTY']._serialized_end=52
  _globals['_TRACEENABLEMESSAGE']._serialized_start=54
  _globals['_TRACEENABLEMESSAGE']._serialized_end=90
  _globals['_TRACEDATAMESSAGE']._serialized_start=92
  _globals['_TRACEDATAMESSAGE']._serialized_end=124
  _globals['_TRACESERVICE']._serialized_start=127
  _globals['_TRACESERVICE']._serialized_end=340
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_i2c/i2c.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10pw_i2c/i2c.proto\x12\x06pw.i2c\"e\n\x0fI2cWriteRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\r\n\x05value\x18\x04 \x01(\x0c\"\x12\n\x10I2cWriteResponse\"h\n\x0eI2cReadRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\x11\n\tread_size\x18\x04 \x01(\r\" \n\x0fI2cReadResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x32\x84\x01\n\x03I2c\x12?\n\x08I2cWrite\x12\x17.pw.i2c.I2cWriteRequest\x1a\x18.pw.i2c.I2cWriteResponse\"\x00\x12<\n\x07I2cRead\x12\x16.pw.i2c.I2cReadRequest\x1a\x17.pw.i2c.I2cReadResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_i2c.i2c_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_I2CWRITEREQUEST']._serialized_start=28
  _globals['_I2CWRITEREQUEST']._serialized_end=129
  _globals['_I2CWRITERESPONSE']._serialized_start=131
  _globals['_I2CWRITERESPONSE']._serialized_end=149
  _globals['_I2CREADREQUEST']._serialized_start=151
  _globals['_I2CREADREQUEST']._serialized_end=255
  _globals['_I2CREADRESPONSE']._serialized_start=257
  _globals['_I2CREADRESPONSE']._serialized_end=289
  _globals['_I2C']._serialized_start=292
  _globals['_I2C']._serialized_end=424
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Command line interface for the Pigweed emulators frontend"""

import argparse
import json
import os
from pathlib import Path
import signal
import subprocess
import sys
import threading

from typing import Any

from pw_emu.core import Error
from pw_emu.frontend import Emulator
from serial import serial_for_url, SerialException
from serial.tools.miniterm import Miniterm, key_description

_TERM_CMD = ['python', '-m', 'serial', '--raw']


def _cmd_gdb_cmds(emu, args: argparse.Namespace) -> None:
    """Run ``gdb`` commands in batch mode."""

    emu.run_gdb_cmds(args.gdb_cmd, executable=args.executable, pause=args.pause)


def _cmd_load(emu: Emulator, args: argparse.Namespace) -> None:
    """Load an executable image via ``gdb`` and start executing it if
    ``--pause`` is not set"""

    args.gdb_cmd = ['load']
    _cmd_gdb_cmds(emu, args)


def _cmd_start(emu: Emulator, args: argparse.Namespace) -> None:
    """Launch the emulator and start executing, unless ``--pause`` is set."""

    if args.runner:
        emu.set_emu(args.runner)

    emu.start(
        target=args.target,
        file=args.file,
        pause=args.pause,
        args=args.args,
        debug=args.debug,
        foreground=args.foreground,
    )


def _get_miniterm(emu: Emulator, chan: str) -> Miniterm:
    chan_type = emu.get_channel_type(chan)
    if chan_type == 'tcp':
        host, port = emu.get_channel_addr(chan)
        url = f'socket://[{host}]:{port}'
    elif chan_type == 'pty':
        url = emu.get_channel_path(chan)
    else:
        raise Error(f'unknown channel type `{chan_type}`')
    ser = serial_for_url(url)
    ser.timeout = 1
    miniterm = Miniterm(ser)
    miniterm.raw = True
    miniterm.set_tx_encoding('UTF-8')
    miniterm.set_rx_encoding('UTF-8')

    quit_key = key_description(miniterm.exit_character)
    menu_key = key_description(miniterm.menu_character)
    help_key = key_description('\x08')
    help_desc = f'Help: {menu_key} followed by {help_key} ---'

    print(f'--- Miniterm on {chan} ---')
    print(f'--- Quit: {quit_key} | Menu: {menu_key} | {help_desc}')

    # On POSIX systems miniterm uses TIOCSTI to "cancel" the TX thread
    # (reading from the console, sending to the serial) which is
    # disabled on Linux kernels > 6.2 see
    # https://github.com/pyserial/pyserial/issues/243
    #
    # On Windows the cancel method does not seem to work either with
    # recent win10 versions.
    #
    # Workaround by terminating the process for exceptions in the read
    # and write threads.
    threading.excepthook = lambda args: signal.raise_signal(signal.SIGTERM)

    return miniterm


def _cmd_run(emu: Emulator, args: argparse.Namespace) -> None:
    """Start the emulator and connect the terminal to a channel. Stop
    the emulator when exiting the terminal."""

    emu.start(
        target=args.target,
        file=args.file,
        pause=True,
        args=args.args,
    )

    ctrl_chans = ['gdb', 'monitor', 'qmp', 'robot']
    if not args.channel:
        for chan in emu.get_channels():
            if chan not in ctrl_chans:
                args.channel = chan
                break
    if not args.channel:
        raise Error(f'only control channels {ctrl_chans} found')

    try:
        miniterm = _get_miniterm(emu, args.channel)
        emu.cont()
        miniterm.start()
        miniterm.join(True)
        print('--- exit ---')
        miniterm.stop()
        miniterm.join()
        miniterm.close()
    except SerialException as err:
        raise Error(f'error connecting to channel `{args.channel}`: {err}')
    finally:
        emu.stop()


def _cmd_restart(emu: Emulator, args: argparse.Namespace) -> None:
    """Restart the emulator and start executing, unless ``--pause`` is set."""

    if emu.running():
        emu.stop()
    _cmd_start(emu, args)


def _cmd_stop(emu: Emulator, _args: argparse.Namespace) -> None:
    """Stop the emulator."""

    emu.stop()


def _cmd_reset(emu: Emulator, _args: argparse.Namespace) -> None:
    """Perform a software reset."""

    emu.reset()


def _cmd_gdb(emu: Emulator, args: argparse.Namespace) -> None:
    """Start a ``gdb`` interactive session."""

    executable = args.executable if args.executable else ""

    signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        cmd = emu.get_gdb_cmd() + [
            '-ex',
            f'target remote {emu.get_gdb_remote()}',
            executable,
        ]
        subprocess.run(cmd)
    finally:
        signal.signal(signal.SIGINT, signal.SIG_DFL)


def _cmd_prop_ls(emu: Emulator, args: argparse.Namespace) -> None:
    """List emulator object properties."""

    props = emu.list_properties(args.path)
    print(json.dumps(props, indent=4))


def _cmd_prop_get(emu: Emulator, args: argparse.Namespace) -> None:
    """Show the emulator's object properties."""

    print(emu.get_property(args.path, args.property))


def _cmd_prop_set(emu: Emulator, args: argparse.Namespace) -> None:
    """Set emulator's object properties."""

    emu.set_property(args.path, args.property, args.value)


def _cmd_term(emu: Emulator, args: argparse.Namespace) -> None:
    """Connect with an interactive terminal to an emulator channel"""

    try:
        miniterm = _get_miniterm(emu, args.channel)
        miniterm.start()
        miniterm.join(True)
        print('--- exit ---')
        miniterm.stop()
        miniterm.join()
        miniterm.close()
    except SerialException as err:
        raise Error(f'error connecting to channel `{args.channel}`: {err}')


def _cmd_resume(emu: Emulator, _args: argparse.Namespace) -> None:
    """Resume the execution of a paused emulator."""

    emu.cont()


def get_parser() -> argparse.ArgumentParser:
    """Command line parser"""

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '-i',
        '--instance',
        help=(
            'Run multiple instances simultaneously by assigning each instance '
            'an ID (default: ``%(default)s``)'
        ),
        type=str,
        metavar='STRING',
        default='default',
    )
    parser.add_argument(
        '-C',
        '--working-dir',
        help=(
            'Absolute path to the working directory '
            '(default: ``%(default)s``)'
        ),
        type=Path,
        default=os.getenv('PW_EMU_WDIR'),
    )
    parser.add_argument(
        '-c',
        '--config',
        help='Absolute path to config file (default: ``%(default)s``)',
        type=str,
        default=None,
    )

    subparsers = parser.add_subparsers(dest='command', required=True)

    def add_cmd(name: str, func: Any) -> argparse.ArgumentParser:
        subparser = subparsers.add_parser(
            name, description=func.__doc__, help=func.__doc__
        )
        subparser.set_defaults(func=func)
        return subparser

    start = add_cmd('start', _cmd_start)
    restart = add_cmd('restart', _cmd_restart)

    for subparser in [start, restart]:
        subparser.add_argument(
            'target',
            type=str,
        )
        subparser.add_argument(
            '--file',
            '-f',
            metavar='FILE',
            help='File to load before starting',
        )
        subparser.add_argument(
            '--runner',
            '-r',
            help='The emulator to use (automatically detected if not set)',
            choices=[None, 'qemu', 'renode'],
            default=None,
        )
        subparser.add_argument(
            '--args',
            '-a',
            help='Options to pass to the emulator',
        )
        subparser.add_argument(
            '--pause',
            '-p',
            action='store_true',
            help='Pause the emulator after starting it',
        )
        subparser.add_argument(
            '--debug',
            '-d',
            action='store_true',
            help='Start the emulator in debug mode',
        )
        subparser.add_argument(
            '--foreground',
            '-F',
            action='store_true',
            help='Start the emulator in foreground mode',
        )

    run = add_cmd('run', _cmd_run)
    run.add_argument(
        'target',
        type=str,
    )
    run.add_argument(
        'file',
        metavar='FILE',
        help='File to load before starting',
    )
    run.add_argument(
        '--args',
        '-a',
        help='Options to pass to the emulator',
    )
    run.add_argument(
        '--channel',
        '-n',
        help='Channel to connect the terminal to',
    )

    stop = add_cmd('stop', _cmd_stop)

    load = add_cmd('load', _cmd_load)
    load.add_argument(
        'executable',
        metavar='FILE',
        help='File to load via ``gdb``',
    )
    load.add_argument(
        '--pause',
        '-p',
        help='Pause the emulator after loading the file',
        action='store_true',
    )
    load.add_argument(
        '--offset',
        '-o',
        metavar='ADDRESS',
        help='Address to load the file at',
    )

    reset = add_cmd('reset', _cmd_reset)

    gdb = add_cmd('gdb', _cmd_gdb)
    gdb.add_argument(
        '--executable',
        '-e',
        metavar='FILE',
        help='File to use for the debugging session',
    )

    prop_ls = add_cmd('prop-ls', _cmd_prop_ls)
    prop_ls.add_argument(
        'path',
        help='Absolute path to the emulator object',
    )

    prop_get = add_cmd('prop-get', _cmd_prop_get)
    prop_get.add_argument(
        'path',
        help='Absolute path to the emulator object',
    )
    prop_get.add_argument(
        'property',
        help='Name of the object property',
    )

    prop_set = add_cmd('prop-set', _cmd_prop_set)
    prop_set.add_argument(
        'path',
        help='Absolute path to the emulator object',
    )
    prop_set.add_argument(
        'property',
        help='Name of the object property',
    )
    prop_set.add_argument(
        'value',
        help='Value to set for the object property',
    )

    gdb_cmds = add_cmd('gdb-cmds', _cmd_gdb_cmds)
    gdb_cmds.add_argument(
        '--pause',
        '-p',
        help='Do not resume execution after running the commands',
        action='store_true',
    )
    gdb_cmds.add_argument(
        '--executable',
        '-e',
        metavar='FILE',
        help='Executable to use while running ``gdb`` commands',
    )
    gdb_cmds.add_argument(
        'gdb_cmd',
        nargs='+',
        help='``gdb`` command to execute',
    )

    term = add_cmd('term', _cmd_term)
    term.add_argument(
        'channel',
        help='Channel name',
    )

    resume = add_cmd('resume', _cmd_resume)

    parser.epilog = f"""commands usage:
        {start.format_usage().strip()}
        {restart.format_usage().strip()}
        {stop.format_usage().strip()}
        {run.format_usage().strip()}
        {load.format_usage().strip()}
        {reset.format_usage().strip()}
        {gdb.format_usage().strip()}
        {prop_ls.format_usage().strip()}
        {prop_get.format_usage().strip()}
        {prop_set.format_usage().strip()}
        {gdb_cmds.format_usage().strip()}
        {term.format_usage().strip()}
        {resume.format_usage().strip()}
    """

    return parser


def main() -> int:
    """Emulators frontend command line interface."""

    args = get_parser().parse_args()
    if not args.working_dir:
        args.working_dir = (
            f'{os.getenv("PW_PROJECT_ROOT")}/.pw_emu/{args.instance}'
        )

    try:
        emu = Emulator(Path(args.working_dir), args.config)
        args.func(emu, args)
    except Error as err:
        print(err)
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Code formatter plugin for GN build files."""

from pathlib import Path
from typing import Final, Iterable, Iterator, Sequence, Tuple

from pw_presubmit.format.core import (
    FileFormatter,
    FormattedFileContents,
    FormatFixStatus,
)


class GnFormatter(FileFormatter):
    """A formatter that runs `gn format` on files."""

    DEFAULT_FLAGS: Final[Sequence[str]] = ()

    def __init__(self, tool_flags: Sequence[str] = DEFAULT_FLAGS, **kwargs):
        super().__init__(**kwargs)
        self.gn_format_flags = list(tool_flags)

    def format_file_in_memory(
        self, file_path: Path, file_contents: bytes
    ) -> FormattedFileContents:
        """Uses ``gn format`` to check the formatting of the requested file.

        The file at ``file_path`` is NOT modified by this check.

        Returns:
            A populated
            :py:class:`pw_presubmit.format.core.FormattedFileContents` that
            contains either the result of formatting the file, or an error
            message.
        """
        # `gn format --dry-run` only signals which files need to be updated,
        # not what the effects are. To get the formatted result, we pipe the
        # file in via stdin and then return the result.
        proc = self.run_tool(
            'gn',
            ['format'] + self.gn_format_flags + ['--stdin'],
            input=file_contents,
        )
        ok = proc.returncode == 0
        return FormattedFileContents(
            ok=ok,
            formatted_file_contents=proc.stdout if ok else b'',
            # `gn format` emits errors over stdout, not stderr.
            error_message=None if ok else proc.stdout.decode(),
        )

    def format_file(self, file_path: Path) -> FormatFixStatus:
        """Formats the provided file in-place using ``gn format``.

        Returns:
            A FormatFixStatus that contains relevant errors/warnings.
        """
        proc = self.run_tool(
            'gn',
            ['format'] + self.gn_format_flags + [file_path],
        )
        ok = proc.returncode == 0
        return FormatFixStatus(
            ok=ok,
            # `gn format` emits errors over stdout, not stderr.
            error_message=None if ok else proc.stdout.decode(),
        )

    def format_files(
        self, paths: Iterable[Path], keep_warnings: bool = True
    ) -> Iterator[Tuple[Path, FormatFixStatus]]:
        """Uses ``gn format`` to fix formatting of the specified files in-place.

        Returns:
            An iterator of ``Path`` and
            :py:class:`pw_presubmit.format.core.FormatFixStatus` pairs for each
            file that was not successfully formatted. If ``keep_warnings`` is
            ``True``, any successful format operations with warnings will also
            be returned.
        """
        # Try to format all files in one `gn` command.
        proc = self.run_tool(
            'gn',
            ['format'] + self.gn_format_flags + list(paths),
        )

        # If there's an error, fall back to per-file formatting to figure out
        # which file has problems.
        if proc.returncode != 0:
            yield from super().format_files(paths, keep_warnings)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Serial reader utilities."""

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
import logging
import os
import platform
import select
import threading
import time
import socket
import subprocess
from typing import (
    Any,
    Callable,
    Iterable,
    Sequence,
    TypeVar,
)
import warnings

import serial

_LOG = logging.getLogger('pw_stream.stream_readers')
_VERBOSE = logging.DEBUG - 1

FrameTypeT = TypeVar('FrameTypeT')


class CancellableReader(ABC):
    """Wraps communication interfaces used for reading incoming data with the
    guarantee that the read request can be cancelled. Derived classes must
    implement the :py:func:`cancel_read()` method.

    Cancelling a read invalidates ongoing and future reads. The
    :py:func:`cancel_read()` method can only be called once.
    """

    def __init__(self, base_obj: Any, *read_args, **read_kwargs):
        """
        Args:
            base_obj: Object that offers a ``read()`` method with optional args
                and kwargs.
            read_args: Arguments for ``base_obj.read()`` function.
            read_kwargs: Keyword arguments for ``base_obj.read()`` function.
        """
        self._base_obj = base_obj
        self._read_args = read_args
        self._read_kwargs = read_kwargs

    def __enter__(self) -> 'CancellableReader':
        return self

    def __exit__(self, *exc_info) -> None:
        self.cancel_read()

    def read(self) -> bytes:
        """Reads bytes that contain parts of or full RPC packets."""
        return self._base_obj.read(*self._read_args, **self._read_kwargs)

    @abstractmethod
    def cancel_read(self) -> None:
        """Cancels a blocking read request and all future reads.

        Can only be called once.
        """


class SelectableReader(CancellableReader):
    """
    Wraps interfaces that work with ``select()`` to signal when data is
    received.

    These interfaces must provide a ``fileno()`` method.
    WINDOWS ONLY: Only sockets that originate from WinSock can be wrapped. File
    objects are not acceptable.
    """

    _STOP_CMD = b'STOP'

    def __init__(self, base_obj: Any, *read_args, **read_kwargs):
        assert hasattr(base_obj, 'fileno')
        if platform.system() == 'Windows' and not isinstance(
            base_obj, socket.socket
        ):
            raise ValueError('Only socket objects are selectable on Windows')
        super().__init__(base_obj, *read_args, **read_kwargs)
        self._cancel_signal_pipe_r_fd, self._cancel_signal_pipe_w_fd = os.pipe()
        self._waiting_for_read_or_cancel_lock = threading.Lock()

    def __exit__(self, *exc_info) -> None:
        self.cancel_read()
        with self._waiting_for_read_or_cancel_lock:
            if self._cancel_signal_pipe_r_fd > 0:
                os.close(self._cancel_signal_pipe_r_fd)
                self._cancel_signal_pipe_r_fd = -1

    def read(self) -> bytes:
        if self._wait_for_read_or_cancel():
            return super().read()
        return b''

    def _wait_for_read_or_cancel(self) -> bool:
        """Returns ``True`` when ready to read."""
        with self._waiting_for_read_or_cancel_lock:
            if self._base_obj.fileno() < 0 or self._cancel_signal_pipe_r_fd < 0:
                # The interface might've been closed already.
                return False
            ready_to_read, _, exception_list = select.select(
                [self._cancel_signal_pipe_r_fd, self._base_obj],
                [],
                [self._base_obj],
            )
            if self._cancel_signal_pipe_r_fd in ready_to_read:
                # A signal to stop the reading process was received.
                os.read(self._cancel_signal_pipe_r_fd, len(self._STOP_CMD))
                os.close(self._cancel_signal_pipe_r_fd)
                self._cancel_signal_pipe_r_fd = -1
                return False

            if exception_list:
                _LOG.error('Error reading interface')
                return False
        return True

    def cancel_read(self) -> None:
        if self._cancel_signal_pipe_w_fd > 0:
            os.write(self._cancel_signal_pipe_w_fd, self._STOP_CMD)
            os.close(self._cancel_signal_pipe_w_fd)
            self._cancel_signal_pipe_w_fd = -1


class SocketReader(SelectableReader):
    """Wraps a socket ``recv()`` function."""

    def __init__(self, base_obj: socket.socket, *read_args, **read_kwargs):
        super().__init__(base_obj, *read_args, **read_kwargs)

    def read(self) -> bytes:
        if self._wait_for_read_or_cancel():
            return self._base_obj.recv(*self._read_args, **self._read_kwargs)
        return b''

    def __exit__(self, *exc_info) -> None:
        self.cancel_read()
        self._base_obj.close()


class SerialReader(CancellableReader):
    """Wraps a :py:class:`serial.Serial` object."""

    def __init__(self, base_obj: serial.Serial, *read_args, **read_kwargs):
        super().__init__(base_obj, *read_args, **read_kwargs)

    def cancel_read(self) -> None:
        self._base_obj.cancel_read()

    def __exit__(self, *exc_info) -> None:
        self.cancel_read()
        self._base_obj.close()


class DataReaderAndExecutor:
    """Reads incoming bytes, data processor that delegates frame handling.

    Executing callbacks in a ``ThreadPoolExecutor`` decouples reading the input
    stream from handling the data. That way, if a handler function takes a
    long time or crashes, this reading thread is not interrupted.
    """

    def __init__(
        self,
        reader: CancellableReader,
        on_read_error: Callable[[Exception], None],
        data_processor: Callable[[bytes], Iterable[FrameTypeT]],
        frame_handler: Callable[[FrameTypeT], None],
        handler_threads: int | None = 1,
    ):
        """Creates the data reader and frame delegator.

        Args:
            reader: Reads incoming bytes from the given transport, blocks until
              data is available or an exception is raised. Otherwise the reader
              will exit.
            on_read_error: Called when there is an error reading incoming bytes.
            data_processor: Processes read bytes and returns a frame-like object
              that the frame_handler can process.
            frame_handler: Handles a received frame.
            handler_threads: The number of threads in the executor pool.
        """

        self._reader = reader
        self._on_read_error = on_read_error
        self._data_processor = data_processor
        self._frame_handler = frame_handler
        self._handler_threads = handler_threads

        self._reader_thread = threading.Thread(target=self._run)
        self._reader_thread_stop = threading.Event()

    def start(self) -> None:
        """Starts the reading process."""
        _LOG.debug('Starting read process')
        self._reader_thread_stop.clear()
        self._reader_thread.start()

    def stop(self) -> None:
        """Stops the reading process.

        This requests that the reading process stop and waits
        for the background thread to exit.
        """
        _LOG.debug('Stopping read process')
        self._reader_thread_stop.set()
        self._reader.cancel_read()
        self._reader_thread.join(30)
        if self._reader_thread.is_alive():
            warnings.warn(
                'Timed out waiting for read thread to terminate.\n'
                'Tip: Use a `CancellableReader` to cancel reads.'
            )

    def _run(self) -> None:
        """Reads raw data in a background thread."""
        with ThreadPoolExecutor(max_workers=self._handler_threads) as executor:
            while not self._reader_thread_stop.is_set():
                try:
                    data = self._reader.read()
                except Exception as exc:  # pylint: disable=broad-except
                    # Don't report the read error if the thread is stopping.
                    # The stream or device backing _read was likely closed,
                    # so errors are expected.
                    if not self._reader_thread_stop.is_set():
                        self._on_read_error(exc)
                    _LOG.debug(
                        'DataReaderAndExecutor thread exiting due to exception',
                        exc_info=exc,
                    )
                    return

                if not data:
                    continue

                _LOG.log(_VERBOSE, 'Read %2d B: %s', len(data), data)

                for frame in self._data_processor(data):
                    executor.submit(self._frame_handler, frame)


def _try_connect(port: int, attempts: int = 10) -> socket.socket:
    """Tries to connect to the specified port up to the given number of times.

    This is helpful when connecting to a process that was started by this
    script. The process may need time to start listening for connections, and
    that length of time can vary. This retries with a short delay rather than
    having to wait for the worst case delay every time.
    """
    timeout_s = 0.001
    while True:
        time.sleep(timeout_s)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(('localhost', port))
            return sock
        except ConnectionRefusedError:
            sock.close()
            attempts -= 1
            if attempts <= 0:
                raise

            timeout_s *= 2


class SocketSubprocess:
    """Executes a subprocess and connects to it with a socket."""

    def __init__(self, command: Sequence, port: int) -> None:
        self._server_process = subprocess.Popen(command, stdin=subprocess.PIPE)
        self.stdin = self._server_process.stdin

        try:
            self.socket: socket.socket = _try_connect(port)  # 🧦
        except:
            self._server_process.terminate()
            self._server_process.communicate()
            raise

    def close(self) -> None:
        try:
            self.socket.close()
        finally:
            self._server_process.terminate()
            self._server_process.communicate()

    def __enter__(self) -> 'SocketSubprocess':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Serve locally-built docs files.

There are essentially four components here:

1. A simple HTTP server that serves docs out of the build directory.

2. JavaScript that tells a doc to refresh itself when it receives a message
   that its source file has been changed. This is injected into each served
   page by #1.

3. A WebSocket server that pushes refresh messages to pages generated by #1
   using the WebSocket client included in #2.

4. A very simple file watcher that looks for changes in the built docs files
   and pushes messages about changed files to #3.
"""

import asyncio
import http.server
import io
import logging
from pathlib import Path
import socketserver
import threading
from tempfile import TemporaryFile
from typing import Callable

from watchdog.events import FileModifiedEvent, FileSystemEventHandler
from watchdog.observers import Observer
import websockets

_LOG = logging.getLogger('pw_docgen.docserver')


def _generate_script(path: str, host: str, port: str) -> bytes:
    """Generate the JavaScript to inject into served docs pages."""
    return f"""<script>
    var connection = null;
    var originFilePath = "{path}";

    function watchForReload() {{
        connection = new WebSocket("ws://{host}:{port}/");
        console.log("Connecting to WebSocket server...");

        connection.onopen = function () {{
            console.log("Connected to WebSocket server");
        }}

        connection.onerror = function () {{
            console.log("WebSocket connection disconnected or failed");
        }}

        connection.onmessage = function (message) {{
            if (message.data === originFilePath) {{
                window.location.reload(true);
            }}
        }}
    }}

    watchForReload();
</script>
</body>
""".encode(
        "utf-8"
    )


class OpenAndInjectScript:
    """A substitute for `open` that injects the refresh handler script.

    Instead of returning a handle to the file you asked for, it returns a
    handle to a temporary file which has been modified. That file will
    disappear as soon as it is `.close()`ed, but that has to be done manually;
    it will not close automatically when exiting scope.

    The instance stores the last path that was opened in `path`.
    """

    def __init__(self, host: str, port: str):
        self.path: str = ""
        self._host = host
        self._port = port

    def __call__(self, path: str, mode: str) -> io.BufferedReader:
        if 'b' not in mode:
            raise ValueError(
                "This should only be used to open files in binary mode."
            )

        content = (
            Path(path)
            .read_bytes()
            .replace(b"</body>", _generate_script(path, self._host, self._port))
        )

        tempfile = TemporaryFile('w+b')
        tempfile.write(content)
        # Let the caller read the file like it's just been opened.
        tempfile.seek(0)
        # Store the path that held the original file.
        self.path = path
        return tempfile  # type: ignore


def _docs_http_server(
    address: str, port: int, path: Path
) -> Callable[[], None]:
    """A simple file system-based HTTP server for built docs."""

    class DocsStaticRequestHandler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=str(path), **kwargs)

        # Disable logs to stdout.
        def log_message(
            self, format: str, *args  # pylint: disable=redefined-builtin
        ) -> None:
            return

    def http_server_thread():
        with socketserver.TCPServer(
            (address, port), DocsStaticRequestHandler
        ) as httpd:
            httpd.serve_forever()

    return http_server_thread


class TaskFinishedException(Exception):
    """Indicates one task has completed successfully."""


class WebSocketConnectionClosedException(Exception):
    """Indicates that the WebSocket connection has been closed."""


class DocsWebsocketRequestHandler:
    """WebSocket server that sends page refresh info to clients.

    Push messages to the message queue to broadcast them to all connected
    clients.
    """

    def __init__(self, address: str = '127.0.0.1', port: int = 8765):
        self._address = address
        self._port = port
        self._connections = set()  # type: ignore
        self._messages: asyncio.Queue = asyncio.Queue()

    async def _register_connection(self, websocket) -> None:
        """Handle client connections and their event loops."""
        self._connections.add(websocket)
        _LOG.info("Client connection established: %s", websocket.id)

        while True:
            try:
                # Run all of these tasks simultaneously. We don't wait for *all*
                # of them to finish -- when one finishes, it raises one of the
                # flow control exceptions to determine what happens next.
                await asyncio.gather(
                    self._send_messages(),
                    self._drop_lost_connection(websocket),
                )
            except TaskFinishedException:
                _LOG.debug("One awaited task finished; iterating event loop.")
            except WebSocketConnectionClosedException:
                _LOG.debug("WebSocket connection closed; ending event loop.")
                return

    async def _drop_lost_connection(self, websocket) -> None:
        """Remove connections to clients with no heartbeat."""
        await asyncio.sleep(1)

        if websocket.closed:
            self._connections.remove(websocket)
            _LOG.info("Client connection dropped: %s", websocket.id)
            raise WebSocketConnectionClosedException

        _LOG.debug("Client connection heartbeat active: %s", websocket.id)
        raise TaskFinishedException

    async def _send_messages(self) -> None:
        """Send the messages in the message queue to all clients.

        Every page change is broadcast to every client. It is up to the client
        to determine whether the contents of a messages means it should refresh.
        This is a pretty easy determination to make though -- the client knows
        its own source file's path, so it just needs to check if the path in the
        message matches it.
        """
        message = await self._messages.get()
        websockets.broadcast(self._connections, message)  # type: ignore # pylint: disable=no-member
        _LOG.info("Sent to %d clients: %s", len(self._connections), message)
        raise TaskFinishedException

    async def _run(self) -> None:
        self._messages = asyncio.Queue()

        async with websockets.serve(  # type: ignore # pylint: disable=no-member
            self._register_connection, self._address, self._port
        ):
            await asyncio.Future()

    def push_message(self, message: str) -> None:
        """Push a message on to the message queue."""
        if len(self._connections) > 0:
            self._messages.put_nowait(message)
            _LOG.info("Pushed to message queue: %s", message)

    def run(self):
        """Run the WebSocket server."""
        asyncio.run(self._run())


class DocsFileChangeEventHandler(FileSystemEventHandler):
    """Handle watched built doc files events."""

    def __init__(self, ws_handler: DocsWebsocketRequestHandler) -> None:
        self._ws_handler = ws_handler

    def on_modified(self, event) -> None:
        if isinstance(event, FileModifiedEvent):
            # Push the path of the modified file to the WebSocket server's
            # message queue.
            path = Path(event.src_path).relative_to(Path.cwd())
            self._ws_handler.push_message(str(path))

        return super().on_modified(event)


class DocsFileChangeObserver(Observer):  # pylint: disable=too-many-ancestors
    """Watch for changes to built docs files."""

    def __init__(
        self, path: str, event_handler: FileSystemEventHandler, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.schedule(event_handler, path, recursive=True)
        _LOG.info("Watching build docs files at: %s", path)


def serve_docs(
    build_dir: Path,
    docs_path: Path,
    address: str = '127.0.0.1',
    port: int = 8000,
    ws_port: int = 8765,
) -> None:
    """Run the docs server.

    This actually spawns three threads, one each for the HTTP server, the
    WebSockets server, and the file watcher.
    """
    docs_path = build_dir.joinpath(docs_path.joinpath('html'))
    http_server_thread = _docs_http_server(address, port, docs_path)

    # The `http.server.SimpleHTTPRequestHandler.send_head` method loads the
    # HTML file from disk, generates and sends headers to the client, then
    # passes the file to the HTTP request handlers. We need to modify the file
    # in the middle of the process, and the only facility we have for doing that
    # is the somewhat distasteful patching of `open`.
    _open_and_inject_script = OpenAndInjectScript(address, str(ws_port))
    setattr(http.server, 'open', _open_and_inject_script)

    websocket_server = DocsWebsocketRequestHandler(address, ws_port)
    event_handler = DocsFileChangeEventHandler(websocket_server)

    threading.Thread(None, websocket_server.run, 'pw_docserver_ws').start()
    threading.Thread(None, http_server_thread, 'pw_docserver_http').start()
    DocsFileChangeObserver(str(docs_path), event_handler).start()

    _LOG.info('Serving docs at http://%s:%d', address, port)


async def ws_client(
    address: str = '127.0.0.1',
    port: int = 8765,
):
    """A simple WebSocket client, useful for testing.

    Run it like this: `asyncio.run(ws_client())`
    """
    async with websockets.connect(f"ws://{address}:{port}") as websocket:  # type: ignore # pylint: disable=no-member
        _LOG.info("Connection ID: %s", websocket.id)
        async for message in websocket:
            _LOG.info("Message received: %s", message)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_digital_io/digital_io.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1epw_digital_io/digital_io.proto\x12\rpw.digital_io\"<\n\x16\x44igitalIoEnableRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\"\x19\n\x17\x44igitalIoEnableResponse\"\\\n\x18\x44igitalIoSetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12,\n\x05state\x18\x02 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState\"\x1b\n\x19\x44igitalIoSetStateResponse\".\n\x18\x44igitalIoGetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\"I\n\x19\x44igitalIoGetStateResponse\x12,\n\x05state\x18\x01 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState**\n\x0e\x44igitalIoState\x12\x0c\n\x08INACTIVE\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x32\xa8\x02\n\tDigitalIo\x12Y\n\x06\x45nable\x12%.pw.digital_io.DigitalIoEnableRequest\x1a&.pw.digital_io.DigitalIoEnableResponse\"\x00\x12_\n\x08SetState\x12\'.pw.digital_io.DigitalIoSetStateRequest\x1a(.pw.digital_io.DigitalIoSetStateResponse\"\x00\x12_\n\x08GetState\x12\'.pw.digital_io.DigitalIoGetStateRequest\x1a(.pw.digital_io.DigitalIoGetStateResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_digital_io.digital_io_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DIGITALIOSTATE']._serialized_start=384
  _globals['_DIGITALIOSTATE']._serialized_end=426
  _globals['_DIGITALIOENABLEREQUEST']._serialized_start=49
  _globals['_DIGITALIOENABLEREQUEST']._serialized_end=109
  _globals['_DIGITALIOENABLERESPONSE']._serialized_start=111
  _globals['_DIGITALIOENABLERESPONSE']._serialized_end=136
  _globals['_DIGITALIOSETSTATEREQUEST']._serialized_start=138
  _globals['_DIGITALIOSETSTATEREQUEST']._serialized_end=230
  _globals['_DIGITALIOSETSTATERESPONSE']._serialized_start=232
  _globals['_DIGITALIOSETSTATERESPONSE']._serialized_end=259
  _globals['_DIGITALIOGETSTATEREQUEST']._serialized_start=261
  _globals['_DIGITALIOGETSTATEREQUEST']._serialized_end=307
  _globals['_DIGITALIOGETSTATERESPONSE']._serialized_start=309
  _globals['_DIGITALIOGETSTATERESPONSE']._serialized_end=382
  _globals['_DIGITALIO']._serialized_start=429
  _globals['_DIGITALIO']._serialized_end=725
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates content related to Pigweed module metadata on pigweed.dev.

This file implements the following pigweed.dev features:

* The `.. pigweed-module::` and `.. pigweed-module-subpage::` directives.
* The auto-generated "Source code" and "Issues" URLs that appear in the site
  nav for each module.

Everything is implemented through the Sphinx Extension API.
"""

from dataclasses import dataclass
import json
import os
import sys
from typing import cast, TypeVar

# We use BeautifulSoup for certain docs rendering features. It may not be
# available in downstream projects. If so, no problem. We fall back to simpler
# docs rendering.
# pylint: disable=import-error
try:
    from bs4 import BeautifulSoup  # type: ignore
    from bs4.element import Tag as HTMLTag  # type: ignore

    bs_enabled = True
except ModuleNotFoundError:
    bs_enabled = False

try:
    import jsonschema  # type: ignore

    jsonschema_enabled = True
except ModuleNotFoundError:
    jsonschema_enabled = False
# pylint: enable=import-error

import docutils
from docutils import nodes
from docutils.nodes import Element
import docutils.statemachine

# pylint: disable=consider-using-from-import
import docutils.parsers.rst.directives as directives  # type: ignore

# pylint: enable=consider-using-from-import
from sphinx.addnodes import document as Document
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.util.docutils import SphinxDirective

from sphinx_design.cards import CardDirective

EnvAttrT = TypeVar('EnvAttrT')


# The module metadata is exposed as a global because it's used as read-only
# data. Opening and reading the metadata file in one of the event handlers
# would cause hundreds of filesystem reads on each build because those event
# handlers fire once for each docs page.
metadata_file = 'docs/module_metadata.json'
schema_file = 'docs/module_metadata_schema.json'
with open(f'{os.environ["PW_ROOT"]}/{schema_file}', 'r') as f:
    schema = json.load(f)
with open(f'{os.environ["PW_ROOT"]}/{metadata_file}', 'r') as f:
    metadata = json.load(f)
# Make sure the metadata matches its schema. Raise an uncaught exception
# if not.
if jsonschema_enabled:
    jsonschema.validate(metadata, schema)


@dataclass
class ParsedBody:
    topnav: str
    body_without_topnav: str


class EnvMetadata:
    """Easier access to the Sphinx `env` for custom metadata.

    You can store things in the Sphinx `env`, which is just a dict. But each
    time you do, you have to handle the possibility that the key you want
    hasn't been set yet, and set it to a default. The `env` is also untyped,
    so you have to cast the value you get to whatever type you expect it to be.

    Or you can use this class to define your metadata keys up front, and just
    access them like: `value = EnvMetadata(env).my_value`

    ... which will handle initializing the value if it hasn't been yet and
    provide you a typed result.
    """

    def __init__(self, env: BuildEnvironment):
        self._env = env

    def _get_env_attr(self, attr: str, default: EnvAttrT) -> EnvAttrT:
        if not hasattr(self._env, attr):
            value: EnvAttrT = default
            setattr(self._env, attr, value)
        else:
            value = getattr(self._env, attr)

        return value

    @property
    def pw_parsed_bodies(self) -> dict[str, ParsedBody]:
        default: dict[str, ParsedBody] = {}
        return self._get_env_attr('pw_module_nav', default)


def get_languages(module_name: str) -> list[str] | None:
    """Returns the list of languages that a module supports.

    Args:
        module_name: The module to look up.

    Returns:
        A list of programming languages that the module supports, or ``None``
        if this has not been defined in ``//docs/module_metadata.json``.
    """
    if module_name not in metadata:
        return None
    if 'languages' not in metadata[module_name]:
        return None
    return metadata[module_name]['languages']


def get_status(module_name: str) -> str:
    """Returns the status of a module.

    Preconditions:
        The status must be defined in ``//docs/module_metadata.json``.

    Args:
        module_name: The module to look up.

    Returns:
        The status of the module as a string.
    """
    if module_name not in metadata:
        sys.exit(f'{module_name} not found in {metadata_file}')
    if 'status' not in metadata[module_name]:
        sys.exit(f'{module_name}.status not found in {metadata_file}')
    return metadata[module_name]['status']


def get_tagline(module_name: str) -> str | None:
    """Returns the tagline for a module.

    Args:
        module_name: The module to look up.

    Returns:
        The module's tagline or ``None`` if no tagline has been defined
        in ``//docs/module_metadata.json``.
    """
    if module_name not in metadata:
        return None
    if 'tagline' not in metadata[module_name]:
        return None
    return metadata[module_name]['tagline']


def get_code_size(module_name: str) -> str | None:
    """Returns the code size impact summary for a module.

    Args:
        module_name: The module to look up.

    Returns:
        The code size impact summary as a string or ``None`` if no summary
        has been defined in ``//docs/module_metadata.json``.
    """
    if module_name not in metadata:
        return None
    if 'size' not in metadata[module_name]:
        return None
    return metadata[module_name]['size']


def status_badge(module_status: str) -> str:
    """Given a module status, return the status badge for rendering."""
    # Use a different badge for each module status value.
    # https://sphinx-design.readthedocs.io/en/latest/badges_buttons.html#badges
    status_to_badge = {
        'stable': 'primary',
        'unstable': 'secondary',
        'experimental': 'warning',
        'deprecated': 'danger',
    }
    badge = status_to_badge[module_status]
    role = f':bdg-{badge}:'
    return role + f'`{module_status.title()}`'


def cs_url(module_name: str) -> str:
    """Return the codesearch URL for the given module."""
    return f'https://cs.opensource.google/pigweed/pigweed/+/main:{module_name}/'


def issues_url(module_name: str) -> str:
    """Returns open issues that mention the given module name."""
    return f'https://issues.pigweed.dev/issues?q={module_name}%20status:open'


def concat_tags(*tag_lists: list[str]) -> list[str]:
    """Given a list of tag lists, return them concat'ed and ready for render."""

    all_tags = tag_lists[0]

    for tag_list in tag_lists[1:]:
        if len(tag_list) > 0:
            all_tags.append(':octicon:`dot-fill`')
            all_tags.extend(tag_list)

    return all_tags


def create_topnav(
    subtitle: str | None,
    extra_classes: list[str] | None = None,
) -> nodes.Node:
    """Create the nodes for the top title and navigation bar."""

    topnav_classes = (
        ['pw-topnav'] + extra_classes if extra_classes is not None else []
    )

    topnav_container = nodes.container(classes=topnav_classes)

    if subtitle:
        subtitle_node = nodes.paragraph(
            classes=['pw-topnav-subtitle'],
            text=subtitle,
        )
        topnav_container += subtitle_node

    return topnav_container


class PigweedModuleDirective(SphinxDirective):
    """Directive registering module metadata, rendering title & info card."""

    required_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {'name': directives.unchanged_required}

    def _try_get_option(self, option: str):
        """Try to get an option by name and raise on failure."""

        try:
            return self.options[option]
        except KeyError:
            raise self.error(f' :{option}: option is required')

    def _maybe_get_option(self, option: str):
        """Try to get an option by name and return None on failure."""
        return self.options.get(option, None)

    def run(self) -> list[nodes.Node]:
        module_name = self._try_get_option('name')
        tagline = get_tagline(module_name)
        status = get_status(module_name)

        status_tags: list[str] = [
            status_badge(status),
        ]

        languages = get_languages(module_name)
        language_tags = []
        if languages:
            for language in languages:
                language_tags.append(f':bdg-info:`{language}`')

        code_size_impact = []

        code_size_text = get_code_size(module_name)
        if code_size_text:
            code_size_impact.append(f'**Code Size Impact:** {code_size_text}')

        # Move the directive content into a section that we can render wherever
        # we want.
        raw_content = cast(list[str], self.content)  # type: ignore
        content = nodes.paragraph()
        self.state.nested_parse(raw_content, 0, content)

        # The card inherits its content from this node's content, which we've
        # already pulled out. So we can replace this node's content with the
        # content we need in the card.
        self.content = docutils.statemachine.StringList(
            concat_tags(status_tags, language_tags, code_size_impact)
        )

        card = CardDirective.create_card(
            inst=self,
            arguments=[],
            options={},
        )

        topbar = create_topnav(
            tagline,
            ['pw-module-index'],
        )

        return [topbar, card, content]


class PigweedModuleSubpageDirective(PigweedModuleDirective):
    """Directive registering module metadata, rendering title & info card."""

    required_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        'name': directives.unchanged_required,
        'nav': directives.unchanged_required,
    }

    def run(self) -> list[nodes.Node]:
        module_name = self._try_get_option('name')
        tagline = get_tagline(module_name)
        # Prepend the module name on sub-pages so that it's very clear what
        # the tagline is referring to.
        tagline = f'{module_name}: {tagline}'

        topbar = create_topnav(
            tagline,
            ['pw-module-subpage'],
        )

        return [topbar]


def _parse_body(body: str) -> ParsedBody:
    """From the `body` HTML, return the topnav and the body without topnav.

    The fundamental idea is this: Our Sphinx directives can only render nodes
    *within* the docutils doc, but we want to elevate the top navbar *outside*
    of that doc into the web theme. Sphinx by itself provides no mechanism for
    this, since it's model looks something like this:

      ┌──────────────────┐
      │ Theme            │
      │  ┌──────────────┐│    When Sphinx builds HTML, the output is plain HTML
      │  │ Sphinx HTML  ││    with a structure defined by docutils. Themes can
      │  │              ││    build *around* that and cascade styles down *into*
      │  │              ││    that HTML, but there's no mechanism in the Sphinx
      │  └──────────────┘│    build to render docutils nodes in the theme.
      └──────────────────┘

    The escape hatch is this:
    - Render things within the Sphinx HTML output (`body`)
    - Use Sphinx theme templates to run code during the final render phase
    - Extract the HTML from the `body` and insert it in the theme via templates

    So this function extracts the things that we rendered in the `body` but
    actually want in the theme (the top navbar), returns them for rendering in
    the template, and returns the `body` with those things removed.
    """
    if not bs_enabled:
        return ParsedBody('', body)

    def _add_class_to_tag(tag: HTMLTag, classname: str) -> None:
        tag['class'] = tag.get('class', []) + [classname]  # type: ignore

    def _add_classes_to_tag(
        tag: HTMLTag, classnames: str | list[str] | None
    ) -> None:
        tag['class'] = tag.get('class', []) + classnames  # type: ignore

    html = BeautifulSoup(body, features='html.parser')

    # Render the doc unchanged, unless it has the module doc topnav
    if (topnav := html.find('div', attrs={'class': 'pw-topnav'})) is None:
        return ParsedBody('', body)

    assert isinstance(topnav, HTMLTag)

    # Find the topnav title and subtitle
    topnav_title = topnav.find('p', attrs={'class': 'pw-topnav-title'})
    topnav_subtitle = topnav.find('p', attrs={'class': 'pw-topnav-subtitle'})
    assert isinstance(topnav_title, HTMLTag)
    assert isinstance(topnav_subtitle, HTMLTag)

    # Find the single `h1` element, the doc's canonical title
    doc_title = html.find('h1')
    assert isinstance(doc_title, HTMLTag)

    topnav_str = ''

    if 'pw-module-index' in topnav['class']:
        # Take the standard Sphinx/docutils title and add topnav styling
        _add_class_to_tag(doc_title, 'pw-topnav-title')
        # Replace the placeholder title in the topnav with the "official" `h1`
        topnav_title.replace_with(doc_title)
        # Promote the subtitle to `h2`
        topnav_subtitle.name = 'h2'
        # We're done mutating topnav; write it to string for rendering elsewhere
        topnav_str = str(topnav)
        # Destroy the instance that was rendered in the document
        topnav.decompose()

    elif 'pw-module-subpage' in topnav['class']:
        # Take the title from the topnav (the module name), promote it to `h1`
        topnav_title.name = 'h1'
        # Add the heading link, but pointed to the module index page
        heading_link = html.new_tag(
            'a',
            attrs={
                'class': ['headerlink'],
                'href': 'docs.html',
                'title': 'Permalink to module index',
            },
        )
        heading_link.string = '#'
        topnav_title.append(heading_link)
        # Promote the subtitle to `h2`
        topnav_subtitle.name = 'h2'
        # We're done mutating topnav; write it to string for rendering elsewhere
        topnav_str = str(topnav)
        # Destroy the instance that was rendered in the document
        topnav.decompose()

    return ParsedBody(topnav_str, str(html))


def setup_parse_body(_app, _pagename, _templatename, context, _doctree):
    def parse_body(body: str) -> ParsedBody:
        return _parse_body(body)

    context['parse_body'] = parse_body


def fix_canonical_url(canonical_url: str | None) -> str | None:
    """Rewrites the canonical URL for `pigweed.dev/*/docs.html` pages.

    Our server is configured to remove `docs.html` from URLs. E.g.
    pigweed.dev/pw_string/docs.html` redirects to `pigweed.dev/pw_string`.
    To improve our SEO, the `<link rel="canonical" href="..."/>` tag in our
    HTML should match the URL that the server provides.

    Args:
        docname:
            Basically the relative path to the doc, except `.rst` is omitted
            from the filename. E.g. `pw_string/docs`.
        canonical_url:
            The default canonical URL that Sphinx has generated for the doc.

    Returns:
        The corrected canonical URL if the page would normally end with
        `docs.html`, otherwise the original canonical URL value unmodified.
    """
    if canonical_url is None or not canonical_url.endswith('/docs.html'):
        return canonical_url
    canonical_url = canonical_url.replace('/docs.html', '/')
    return canonical_url


def on_html_page_context(
    app: Sphinx,  # pylint: disable=unused-argument
    docname: str,  # pylint: disable=unused-argument
    templatename: str,  # pylint: disable=unused-argument
    context: dict[str, str | None] | None,
    doctree: Document,  # pylint: disable=unused-argument
) -> None:
    """Handles modifications to HTML page metadata, e.g. canonical URLs.

    Args:
        docname:
            Basically the relative path to the doc, except `.rst` is omitted
            from the filename. E.g. `pw_string/docs`.
        context:
            A dict containing the HTML page's metadata.

    Returns:
        None. Modifications happen to the HTML metadata in-place.
    """
    canonical_url_key = 'pageurl'
    if context is None or canonical_url_key not in context:
        return
    canonical_url = context[canonical_url_key]
    context[canonical_url_key] = fix_canonical_url(canonical_url)


def add_links(module_name: str, toctree: Element) -> None:
    """Adds source code and issues URLs to a module's table of contents tree.

    This function is how we auto-generate the source code and issues URLs
    that appear for each module in the pigweed.dev site nav.

    Args:
        module_name:
            The Pigweed module that we're creating links for.
        toctree:
            The table of contents tree from that module's homepage.

    Returns:
        `None`. `toctree` is modified in-place.
    """
    src = ('Source code', cs_url(module_name))
    issues = ('Issues', issues_url(module_name))
    # Maintenance tip: the trick here is to create the `toctree` the same way
    # that Sphinx generates it. When in doubt, enable logging in this file,
    # manually modify the `.. toctree::` directive on a module's homepage, log
    # out `toctree` from somewhere in this script (you should see an XML-style
    # node), and then just make sure your code modifies the `toctree` the same
    # way that Sphinx generates it.
    toctree['entries'] += [src, issues]
    toctree['rawentries'] += [src[0], issues[0]]


def find_first_toctree(doctree: Document) -> Element | None:
    """Finds the first `toctree` (table of contents tree) node in a `Document`.

    Args:
        doctree:
            The content of a doc, represented as a tree of Docutils nodes.

    Returns:
        The first `toctree` node found in `doctree` or `None` if none was
        found.
    """
    for node in doctree.traverse(nodes.Element):
        if node.tagname == 'toctree':
            return node
    return None


def parse_module_name(docname: str) -> str:
    """Extracts a Pigweed module name from a Sphinx docname.

    Preconditions:
        `docname` is assumed to start with `pw_`. I.e. the docs are assumed to
        have a flat directory structure, where the first directory is the name
        of a Pigweed module.

    Args:
        docname:
            Basically the relative path to the doc, except `.rst` is omitted
            from the filename. E.g. `pw_string/docs`.

    Returns:
        Just the Pigweed module name, e.g. `pw_string`.
    """
    tokens = docname.split('/')
    return tokens[0]


def on_doctree_read(app: Sphinx, doctree: Document) -> None:
    """Event handler that enables manipulating a doc's Docutils tree.

    Sphinx fires this listener after it has parsed a doc's reStructuredText
    into a tree of Docutils nodes. The listener fires once for each doc that's
    processed.

    In general, this stage of the Sphinx event lifecycle can only be used for
    content changes that do not affect the Sphinx build environment [1]. For
    example, creating a `toctree` node at this stage does not work, but
    inserting links into a pre-existing `toctree` node is OK.

    Args:
        app:
            Our Sphinx docs build system.
        doctree:
            The doc content, structured as a tree.

    Returns:
        `None`. The main modifications happen in-place in `doctree`.

    [1] See link in `on_source_read()`
    """
    docname = app.env.docname
    if not is_module_homepage(docname):
        return
    toctree = find_first_toctree(doctree)
    if toctree is None:
        # `add_toctree_to_module_homepage()` should ensure that every
        # `pw_*/docs.rst` file has a `toctree` node but if something went wrong
        # then we should bail.
        sys.exit(f'[module_metadata.py] error: toctree missing in {docname}')
    module_name = parse_module_name(docname)
    add_links(module_name, toctree)


def is_module_homepage(docname: str) -> bool:
    """Determines if a doc is a module homepage.

    Any doc that matches the pattern `pw_*/docs.rst` is considered a module
    homepage. Watch out for the false positive of `pw_*/*/docs.rst`.

    Preconditions:
        `docname` is assumed to start with `pw_`. I.e. the docs are assumed to
        have a flat directory structure, where the first directory is the name
        of a Pigweed module.

    Args:
        docname:
            Basically the relative path to the doc, except `.rst` is omitted
            from the filename.

    Returns:
        `True` if the doc is a module homepage, else `False`.
    """
    tokens = docname.split('/')
    if len(tokens) != 2:
        return False
    if not tokens[0].startswith('pw_'):
        return False
    if tokens[1] != 'docs':
        return False
    return True


def add_toctree_to_module_homepage(docname: str, source: str) -> str:
    """Appends an empty `toctree` to a module homepage.

    Note that this function only needs to create the `toctree` node; it doesn't
    need to fully populate the `toctree`. Inserting links later via the more
    ergonomic Docutils API works fine.

    Args:
        docname:
            Basically the relative path to `source`, except `.rst` is omitted
            from the filename.
        source:
            The reStructuredText source code of `docname`.

    Returns:
        For module homepages that did not already have a `toctree`, the
        original contents of `source` plus an empty `toctree` is returned.
        For all other cases, the original contents of `source` are returned
        with no modification.
    """
    # Don't do anything if the page is not a module homepage, i.e. its
    # `docname` doesn't match the pattern `pw_*`/docs`.
    if not is_module_homepage(docname):
        return source
    # Don't do anything if the module homepage already has a `toctree`.
    if '.. toctree::' in source:
        return source
    # Append an empty `toctree` to the content.
    # yapf: disable
    return (
        f'{source}\n\n'
        '.. toctree::\n'
        '   :hidden:\n'
        '   :maxdepth: 1\n'
    )
    # yapf: enable
    # Python formatting (yapf) is disabled in the return statement because the
    # formatter tries to change it to a less-readable single line string.


# inclusive-language: disable
def on_source_read(
    app: Sphinx,  # pylint: disable=unused-argument
    docname: str,
    source: list[str],
) -> None:
    """Event handler that enables manipulating a doc's reStructuredText.

    Sphinx fires this event early in its event lifecycle [1], before it has
    converted a doc's reStructuredText (reST) into a tree of Docutils nodes.
    The listener fires once for each doc that's processed.

    This is the place to make docs changes that have to propagate across the
    site. Take our use case of adding a link in the site nav to each module's
    source code. To do this we need a `toctree` (table of contents tree) node
    on each module's homepage; the `toctree` is where we insert the source code
    link. If we try to dynamically insert the `toctree` node via the Docutils
    API later in the event lifecycle, e.g. during the `doctree-read` event, we
    have to do a bunch of complex and fragile logic to make the Sphinx build
    environment [2] aware of the new node. It's simpler and more reliable to
    just insert a `.. toctree::` directive into the doc source before Sphinx
    has processed the doc and then let Sphinx create its build environment as
    it normally does. We just have to make sure the reStructuredText we're
    injecting into the content is syntactically correct.

    Args:
        app:
            Our Sphinx docs build system.
        docname:
            Basically the relative path to `source`, except `.rst` is omitted
            from the filename.
        source:
            The reStructuredText source code of `docname`.

    Returns:
        None. `source` is modified in-place.

    [1] www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx-core-events
    [2] www.sphinx-doc.org/en/master/extdev/envapi.html
    """
    # inclusive-language: enable
    # If a module homepage doesn't have a `toctree`, add one.
    source[0] = add_toctree_to_module_homepage(docname, source[0])


def setup(app: Sphinx) -> dict[str, bool]:
    """Hooks the extension into our Sphinx docs build system.

    This runs only once per docs build.

    Args:
        app:
            Our Sphinx docs build system.

    Returns:
        A dict that provides Sphinx info about our extension.
    """
    # Register the `.. pigweed-module::` and `.. pigweed-module-subpage::`
    # directives that are used on `pw_*/*.rst` pages.
    app.add_directive('pigweed-module', PigweedModuleDirective)
    app.add_directive('pigweed-module-subpage', PigweedModuleSubpageDirective)
    # inclusive-language: disable
    # Register the Sphinx event listeners that automatically generate content
    # for `pw_*/*.rst` pages:
    # www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx-core-events
    # inclusive-language: enable
    app.connect('source-read', on_source_read)
    app.connect('doctree-read', on_doctree_read)
    app.connect('html-page-context', on_html_page_context)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Configure Visual Studio Code (VSC) for Pigweed projects.

VSC recognizes three sources of configurable settings:

1. Project settings, stored in {project root}/.vscode/settings.json
2. Workspace settings, stored in (workspace root)/.vscode/settings.json;
   a workspace is a collection of projects/repositories that are worked on
   together in a single VSC instance
3. The user's personal settings, which are stored somewhere in the user's home
   directory, and are applied to all instances of VSC

This provides three levels of settings cascading:

    Workspace <- Project <- User

... where the arrow indicates the ability to override.

Out of these three, only project settings are useful to Pigweed projects. User
settings are essentially global and outside the scope of Pigweed. Workspaces are
seldom used and don't work well with the Pigweed directory structure.

Nonetheless, we want a three-tiered settings structure for Pigweed projects too:

A. Default settings provided by Pigweed, configuring VSC to use IDE features
B. Project-level overrides that downstream projects may define
C. User-level overrides that individual users may define

We accomplish all of that with only the project settings described in #1 above.

Default settings are defined in this module. Project settings can be defined in
.vscode/pw_project_settings.json and should be checked into the repository. User
settings can be defined in .vscode/pw_user_settings.json and should not be
checked into the repository. None of these settings have any effect until they
are merged into VSC's settings (.vscode/settings.json) via the functions in this
module. Those resulting settings are system-specific and should also not be
checked into the repository.

We provide the same structure to both tasks and extensions as well. Defaults
are provided by Pigweed, can be augmented or overridden at the project level
with .vscode/pw_project_tasks.json and .vscode/pw_project_extensions.json,
can be augmented or overridden by an individual developer with
.vscode/pw_user_tasks.json and .vscode/pw_user.extensions.json, and none of
this takes effect until they are merged into VSC's active settings files
(.vscode/tasks.json and .vscode/extensions.json) by running the appropriate
command.
"""

from __future__ import annotations

# TODO(chadnorvell): Import collections.OrderedDict when we don't need to
# support Python 3.8 anymore.
from enum import Enum
from pathlib import Path
import shutil
import subprocess
from typing import Any, Callable, OrderedDict

from pw_cli.env import pigweed_environment

from pw_ide.cpp import ClangdSettings, CppIdeFeaturesState

from pw_ide.editors import (
    EditorSettingsDict,
    EditorSettingsManager,
    EditorSettingsTypesWithDefaults,
    Json5FileFormat,
)

from pw_ide.python import PythonPaths
from pw_ide.settings import PigweedIdeSettings

env = pigweed_environment()


def _local_clangd_settings(ide_settings: PigweedIdeSettings) -> dict[str, Any]:
    """VSC settings for running clangd with Pigweed paths."""
    clangd_settings = ClangdSettings(ide_settings)
    return {
        'clangd.path': str(clangd_settings.clangd_path),
        'clangd.arguments': clangd_settings.arguments,
    }


def _local_python_settings() -> dict[str, Any]:
    """VSC settings for finding the Python virtualenv."""
    paths = PythonPaths()
    return {
        'python.defaultInterpreterPath': str(paths.interpreter),
    }


# The order is preserved despite starting with a plain dict because in Python
# 3.6+, plain dicts are actually ordered as an implementation detail. This could
# break on interpreters other than CPython, or if the implementation changes in
# the future. However, for now, this is much more readable than the more robust
# alternative of instantiating with a list of tuples.
_DEFAULT_SETTINGS: EditorSettingsDict = OrderedDict(
    {
        "cmake.format.allowOptionalArgumentIndentation": True,
        "editor.detectIndentation": False,
        "editor.rulers": [80],
        "editor.tabSize": 2,
        "files.associations": OrderedDict({"*.inc": "cpp"}),
        "files.exclude": OrderedDict(
            {
                "**/*.egg-info": True,
                "**/.mypy_cache": True,
                "**/__pycache__": True,
                ".cache": True,
                ".cipd": True,
                ".environment": True,
                ".presubmit": True,
                ".pw_ide": True,
                ".pw_ide.user.yaml": True,
                "bazel-bin": True,
                "bazel-out": True,
                "bazel-pigweed": True,
                "bazel-testlogs": True,
                "environment": True,
                "node_modules": True,
                "out": True,
            }
        ),
        "files.insertFinalNewline": True,
        "files.trimTrailingWhitespace": True,
        "search.useGlobalIgnoreFiles": True,
        "npm.autoDetect": "off",
        "C_Cpp.intelliSenseEngine": "disabled",
        "[cpp]": OrderedDict(
            {"editor.defaultFormatter": "llvm-vs-code-extensions.vscode-clangd"}
        ),
        "python.analysis.diagnosticSeverityOverrides": OrderedDict(
            # Due to our project structure, the linter spuriously thinks we're
            # shadowing system modules any time we import them. This disables
            # that check.
            {"reportShadowedImports": "none"}
        ),
        # The "strict" mode is much more strict than what we currently enforce.
        "python.analysis.typeCheckingMode": "basic",
        # Restrict the search for Python files to the locations we expect to
        # have Python files. This minimizes the time & RAM the LSP takes to
        # parse the project.
        "python.analysis.include": ["pw_*/py/**/*"],
        "python.terminal.activateEnvironment": False,
        "python.testing.unittestEnabled": True,
        "[python]": OrderedDict({"editor.tabSize": 4}),
        "typescript.tsc.autoDetect": "off",
        "[gn]": OrderedDict({"editor.defaultFormatter": "msedge-dev.gnls"}),
        "[proto3]": OrderedDict(
            {"editor.defaultFormatter": "zxh404.vscode-proto3"}
        ),
        "[restructuredtext]": OrderedDict({"editor.tabSize": 3}),
    }
)

_DEFAULT_TASKS: EditorSettingsDict = OrderedDict(
    {
        "version": "2.0.0",
        "tasks": [
            {
                "type": "process",
                "label": "Pigweed: Format",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw format --fix'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: Presubmit",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw presubmit'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Set Python Virtual Environment",
                "command": "${command:python.setInterpreter}",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Restart Python Language Server",
                "command": "${command:python.analysis.restartLanguageServer}",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Restart C++ Language Server",
                "command": "${command:clangd.restart}",
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: Sync IDE",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide sync'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: Current C++ Target Toolchain",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide cpp'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: List C++ Target Toolchains",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide cpp --list'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": (
                    "Pigweed: Change C++ Target Toolchain "
                    "without LSP restart"
                ),
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide cpp --set ${input:availableTargetToolchains}'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Set C++ Target Toolchain",
                "dependsOrder": "sequence",
                "dependsOn": [
                    "Pigweed: Change C++ Target Toolchain without LSP restart",
                    "Pigweed: Restart C++ Language Server",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
        ],
    }
)

_DEFAULT_EXTENSIONS: EditorSettingsDict = OrderedDict(
    {
        "recommendations": [
            "llvm-vs-code-extensions.vscode-clangd",
            "ms-python.mypy-type-checker",
            "ms-python.python",
            "ms-python.pylint",
            "npclaudiu.vscode-gn",
            "msedge-dev.gnls",
            "zxh404.vscode-proto3",
            "josetr.cmake-language-support-vscode",
        ],
        "unwantedRecommendations": [
            "ms-vscode.cpptools",
            "persidskiy.vscode-gnformat",
            "lextudio.restructuredtext",
        ],
    }
)

_DEFAULT_LAUNCH: EditorSettingsDict = OrderedDict(
    {
        "version": "0.2.0",
        "configurations": [],
    }
)


def _default_settings(
    pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return OrderedDict(
        {
            **_DEFAULT_SETTINGS,
            **_local_clangd_settings(pw_ide_settings),
            **_local_python_settings(),
        }
    )


def _default_settings_no_side_effects(
    _pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return OrderedDict(
        {
            **_DEFAULT_SETTINGS,
        }
    )


def _default_tasks(
    pw_ide_settings: PigweedIdeSettings,
    state: CppIdeFeaturesState | None = None,
) -> EditorSettingsDict:
    if state is None:
        state = CppIdeFeaturesState(pw_ide_settings)

    inputs = [
        {
            "type": "pickString",
            "id": "availableTargetToolchains",
            "description": "Available target toolchains",
            "options": sorted(list(state.targets)),
        }
    ]

    return OrderedDict(**_DEFAULT_TASKS, inputs=inputs)


def _default_extensions(
    _pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return _DEFAULT_EXTENSIONS


def _default_launch(
    _pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return _DEFAULT_LAUNCH


DEFAULT_SETTINGS_PATH: Callable[[PigweedIdeSettings], Path] = (
    lambda settings: settings.workspace_root / '.vscode'
)


class VscSettingsType(Enum):
    """Visual Studio Code settings files.

    VSC supports editor settings (``settings.json``), recommended
    extensions (``extensions.json``), tasks (``tasks.json``), and
    launch/debug configurations (``launch.json``).
    """

    SETTINGS = 'settings'
    TASKS = 'tasks'
    EXTENSIONS = 'extensions'
    LAUNCH = 'launch'

    @classmethod
    def all(cls) -> list[VscSettingsType]:
        return list(cls)


class VscSettingsManager(EditorSettingsManager[VscSettingsType]):
    """Manages all settings for Visual Studio Code."""

    default_settings_dir = DEFAULT_SETTINGS_PATH
    file_format = Json5FileFormat()

    types_with_defaults: EditorSettingsTypesWithDefaults = {
        VscSettingsType.SETTINGS: _default_settings,
        VscSettingsType.TASKS: _default_tasks,
        VscSettingsType.EXTENSIONS: _default_extensions,
        VscSettingsType.LAUNCH: _default_launch,
    }


class VscSettingsManagerNoSideEffects(EditorSettingsManager[VscSettingsType]):
    """This is like VscSettingsManager, but optimized for unit testing."""

    default_settings_dir = DEFAULT_SETTINGS_PATH
    file_format = Json5FileFormat()

    types_with_defaults: EditorSettingsTypesWithDefaults = {
        VscSettingsType.SETTINGS: _default_settings_no_side_effects,
        VscSettingsType.TASKS: _default_tasks,
        VscSettingsType.EXTENSIONS: _default_extensions,
        VscSettingsType.LAUNCH: _default_launch,
    }


def build_extension(pw_root: Path):
    """Build the VS Code extension."""

    license_path = pw_root / 'LICENSE'
    icon_path = pw_root.parent / 'icon.png'

    vsc_ext_path = pw_root / 'pw_ide' / 'ts' / 'pigweed-vscode'
    out_path = vsc_ext_path / 'out'
    dist_path = vsc_ext_path / 'dist'
    temp_license_path = vsc_ext_path / 'LICENSE'
    temp_icon_path = vsc_ext_path / 'icon.png'

    shutil.rmtree(out_path, ignore_errors=True)
    shutil.rmtree(dist_path, ignore_errors=True)
    shutil.copy(license_path, temp_license_path)
    shutil.copy(icon_path, temp_icon_path)

    try:
        subprocess.run(['npm', 'install'], check=True, cwd=vsc_ext_path)
        subprocess.run(['npm', 'run', 'compile'], check=True, cwd=vsc_ext_path)
        subprocess.run(['vsce', 'package'], check=True, cwd=vsc_ext_path)
    except subprocess.CalledProcessError as e:
        raise e
    finally:
        temp_license_path.unlink()
        temp_icon_path.unlink()



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Utils to decode logs."""

import logging
import warnings

from pw_log.log_decoder import LogStreamDecoder
from pw_log.proto import log_pb2
import pw_rpc
import pw_status

_LOG = logging.getLogger(__name__)


class LogStreamHandler:
    """Handles an RPC Log Stream.

    Args:
        rpcs: RPC services to request RPC Log Streams.
        decoder: LogStreamDecoder
    """

    def __init__(
        self, rpcs: pw_rpc.client.Services, decoder: LogStreamDecoder
    ) -> None:
        self.rpcs = rpcs
        self._decoder = decoder

    def listen_to_logs(self) -> None:
        warnings.warn(
            'listen_to_logs is deprecated; call start_logging() instead',
            DeprecationWarning,
        )
        self.start_logging()

    def start_logging(self) -> None:
        """Requests logs to be streamed over the pw.log.Logs.Listen RPC."""
        self.rpcs.pw.log.Logs.Listen.invoke(
            on_next=self._on_log_entries,
            on_completed=lambda _, status: self.handle_log_stream_completed(
                status
            ),
            on_error=lambda _, error: self.handle_log_stream_error(error),
        )

    def _on_log_entries(self, _, log_entries_proto: log_pb2.LogEntries) -> None:
        self._decoder.parse_log_entries_proto(log_entries_proto)

    def handle_log_stream_error(self, error: pw_status.Status) -> None:
        """Resets the log stream RPC on error to avoid losing logs.

        Override this function to change default behavior.
        """
        _LOG.error(
            'Log stream error: %s from source %s',
            error,
            self.source_name,
        )
        # Only re-request logs if the RPC was not cancelled by the client.
        if error != pw_status.Status.CANCELLED:
            self.start_logging()

    def handle_log_stream_completed(self, status: pw_status.Status) -> None:
        """Resets the log stream RPC on completed to avoid losing logs.

        Override this function to change default behavior.
        """
        _LOG.debug(
            'Log stream completed with status: %s for source: %s',
            status,
            self.source_name,
        )
        self.start_logging()

    @property
    def source_name(self) -> str:
        return self._decoder.source_name



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class RiscvCpuState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MEPC_FIELD_NUMBER: builtins.int
    MCAUSE_FIELD_NUMBER: builtins.int
    MSTATUS_FIELD_NUMBER: builtins.int
    MTVAL_FIELD_NUMBER: builtins.int
    RA_FIELD_NUMBER: builtins.int
    SP_FIELD_NUMBER: builtins.int
    T0_FIELD_NUMBER: builtins.int
    T1_FIELD_NUMBER: builtins.int
    T2_FIELD_NUMBER: builtins.int
    FP_FIELD_NUMBER: builtins.int
    S1_FIELD_NUMBER: builtins.int
    A0_FIELD_NUMBER: builtins.int
    A1_FIELD_NUMBER: builtins.int
    A2_FIELD_NUMBER: builtins.int
    A3_FIELD_NUMBER: builtins.int
    A4_FIELD_NUMBER: builtins.int
    A5_FIELD_NUMBER: builtins.int
    A6_FIELD_NUMBER: builtins.int
    A7_FIELD_NUMBER: builtins.int
    S2_FIELD_NUMBER: builtins.int
    S3_FIELD_NUMBER: builtins.int
    S4_FIELD_NUMBER: builtins.int
    S5_FIELD_NUMBER: builtins.int
    S6_FIELD_NUMBER: builtins.int
    S7_FIELD_NUMBER: builtins.int
    S8_FIELD_NUMBER: builtins.int
    S9_FIELD_NUMBER: builtins.int
    S10_FIELD_NUMBER: builtins.int
    S11_FIELD_NUMBER: builtins.int
    T3_FIELD_NUMBER: builtins.int
    T4_FIELD_NUMBER: builtins.int
    T5_FIELD_NUMBER: builtins.int
    T6_FIELD_NUMBER: builtins.int
    mepc: builtins.int
    mcause: builtins.int
    mstatus: builtins.int
    mtval: builtins.int
    ra: builtins.int
    sp: builtins.int
    t0: builtins.int
    t1: builtins.int
    t2: builtins.int
    fp: builtins.int
    s1: builtins.int
    a0: builtins.int
    a1: builtins.int
    a2: builtins.int
    a3: builtins.int
    a4: builtins.int
    a5: builtins.int
    a6: builtins.int
    a7: builtins.int
    s2: builtins.int
    s3: builtins.int
    s4: builtins.int
    s5: builtins.int
    s6: builtins.int
    s7: builtins.int
    s8: builtins.int
    s9: builtins.int
    s10: builtins.int
    s11: builtins.int
    t3: builtins.int
    t4: builtins.int
    t5: builtins.int
    t6: builtins.int
    def __init__(
        self,
        *,
        mepc: builtins.int | None = ...,
        mcause: builtins.int | None = ...,
        mstatus: builtins.int | None = ...,
        mtval: builtins.int | None = ...,
        ra: builtins.int | None = ...,
        sp: builtins.int | None = ...,
        t0: builtins.int | None = ...,
        t1: builtins.int | None = ...,
        t2: builtins.int | None = ...,
        fp: builtins.int | None = ...,
        s1: builtins.int | None = ...,
        a0: builtins.int | None = ...,
        a1: builtins.int | None = ...,
        a2: builtins.int | None = ...,
        a3: builtins.int | None = ...,
        a4: builtins.int | None = ...,
        a5: builtins.int | None = ...,
        a6: builtins.int | None = ...,
        a7: builtins.int | None = ...,
        s2: builtins.int | None = ...,
        s3: builtins.int | None = ...,
        s4: builtins.int | None = ...,
        s5: builtins.int | None = ...,
        s6: builtins.int | None = ...,
        s7: builtins.int | None = ...,
        s8: builtins.int | None = ...,
        s9: builtins.int | None = ...,
        s10: builtins.int | None = ...,
        s11: builtins.int | None = ...,
        t3: builtins.int | None = ...,
        t4: builtins.int | None = ...,
        t5: builtins.int | None = ...,
        t6: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["a0", b"a0", "a1", b"a1", "a2", b"a2", "a3", b"a3", "a4", b"a4", "a5", b"a5", "a6", b"a6", "a7", b"a7", "fp", b"fp", "mcause", b"mcause", "mepc", b"mepc", "mstatus", b"mstatus", "mtval", b"mtval", "ra", b"ra", "s1", b"s1", "s10", b"s10", "s11", b"s11", "s2", b"s2", "s3", b"s3", "s4", b"s4", "s5", b"s5", "s6", b"s6", "s7", b"s7", "s8", b"s8", "s9", b"s9", "sp", b"sp", "t0", b"t0", "t1", b"t1", "t2", b"t2", "t3", b"t3", "t4", b"t4", "t5", b"t5", "t6", b"t6"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["a0", b"a0", "a1", b"a1", "a2", b"a2", "a3", b"a3", "a4", b"a4", "a5", b"a5", "a6", b"a6", "a7", b"a7", "fp", b"fp", "mcause", b"mcause", "mepc", b"mepc", "mstatus", b"mstatus", "mtval", b"mtval", "ra", b"ra", "s1", b"s1", "s10", b"s10", "s11", b"s11", "s2", b"s2", "s3", b"s3", "s4", b"s4", "s5", b"s5", "s6", b"s6", "s7", b"s7", "s8", b"s8", "s9", b"s9", "sp", b"sp", "t0", b"t0", "t1", b"t1", "t2", b"t2", "t3", b"t3", "t4", b"t4", "t5", b"t5", "t6", b"t6"]) -> None: ...

global___RiscvCpuState = RiscvCpuState

@typing_extensions.final
class SnapshotCpuStateOverlay(google.protobuf.message.Message):
    """This message overlays the pw.snapshot.Snapshot proto. It's valid to encode
    this message to the same sink that a Snapshot proto is being written to.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RISCV_CPU_STATE_FIELD_NUMBER: builtins.int
    @property
    def riscv_cpu_state(self) -> global___RiscvCpuState: ...
    def __init__(
        self,
        *,
        riscv_cpu_state: global___RiscvCpuState | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["riscv_cpu_state", b"riscv_cpu_state"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["riscv_cpu_state", b"riscv_cpu_state"]) -> None: ...

global___SnapshotCpuStateOverlay = SnapshotCpuStateOverlay
